/*****************************************************************

  grip.c

  Copyright (c) 1998, 1999 by Mike Oliphant - oliphant@ling.ed.ac.uk

    http://www.ling.ed.ac.uk/~oliphant/grip

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include "grip.h"
#include "xpm.h"
#include "config.h"
#include "parsecfg.h"

void ShutDown(GtkWidget *widget, GdkEvent *event, gpointer data);
gboolean FileExists(char *filename);
void UpdateGTK(void);
void Busy(void);
void UnBusy(void);
void DisplayMsg(GtkWidget *widget,GdkEventButton *event,gpointer data);
void MinMax(void);
void ToggleVol(void);
void ToggleProg(void);
void PlaylistChanged(void);
void ToggleTrackEdit(void);
void SetTitle(char *title);
void SetArtist(char *artist);
void SetCDDBGenre(int genre);
void SaveDiscInfo(void);
void TitleEditChanged(void);
void ArtistEditChanged(void);
void TrackEditChanged(void);
void CDDBGenreChanged(void);
gint TimeOut(gpointer data);
char *UnderScore(char *str);
void ParseFileFmt(char *instr,char *outstr,int track,int startsec,int endsec);
void KillChild(void);
void SetVolume(GtkWidget *widget,gpointer data);
void NextTrack(void);
void PrevTrack(void);
void InitProgram(int num_tracks,int mode);
void ToggleLoop(void);
void ChangePlayMode(void);
void FastFwdCB(GtkWidget *widget,gpointer data);
void FastFwd(void);
void RewindCB(GtkWidget *widget,gpointer data);
void Rewind(void);
void NextDisc(void);
void StopPlay(void);
void EjectDisc(void);
void PlayTrackCB(GtkWidget *widget,gpointer data);
void PlayTrack(int track);
void ChangeTimeMode(void);
void CheckNewDisc(void);
void UpdateDisplay(void);
gboolean CheckTracks(void);
void SetCurrentTrack(int track);
gboolean TrackIsChecked(int track);
void ToggleChecked(int track);
void SetChecked(int track,gboolean checked);
void ClickColumn(GtkWidget *widget,gint column,gpointer data);
void CListButtonPressed(GtkWidget *widget,GdkEventButton *event,gpointer data);
void UnSelectRow(GtkWidget *widget,gint row,gint column,
		 GdkEventButton *event,gpointer data);
void SelectRow(GtkWidget *widget,gint row,gint column,
	       GdkEventButton *event,gpointer data);
void CDDBToggle(void);
void LookupDisc(gboolean manual);
void DoLookup(void);
gboolean CDDBLookupDisc(CDDBServer *server);
void SubmitEntryCB(void);
void SubmitEntry(void);
void UpdateTracks(void);
void ChangeStrVal(GtkWidget *widget,gpointer data);
GtkWidget *MakeStrEntry(char *name,int num,int len,gboolean editable);
void ChangeIntVal(GtkWidget *widget,gpointer data);
GtkWidget *MakeNumEntry(char *name,int num,int len);
void ChangeBoolVal(GtkWidget *widget,gpointer data);
GtkWidget *MakeCheckButton(char *name,int num);
GtkWidget *MakeNewPage(char *name);
void MakeTrackPage(void);
void ChopCR(char *str);
void DoLoadConfig(void);
void DoSaveConfig(void);
void MakeConfigPage(void);
void Homepage(void);
void MakeHelpPage(void);
void MakeAboutPage(void);
GtkTooltips *MakeToolTip(void);
GtkWidget *MakeEditBox(void);
GtkWidget *MakePlayOpts(void);
GtkWidget *MakeControls(void);
GtkWidget *ImageButton(char **xpm);
GtkWidget *ImageToggleButton(char **xpm, gint state);
GtkWidget *Loadxpm(char **xpm);
GtkWidget *NewBlankPixmap(void);
void CopyPixmap(GtkPixmap *src,GtkPixmap *dest);
void Usage(void);
void ParseArgs(int numargs,char *args[]);
void ShuffleTracks(int num_tracks);

/* Rip-related functions */

#ifndef GRIPCD

#ifdef MP3DB
void DBScan(void);
void AddSQLEntry(int track);
#endif

void CheckDupNames(void);
void UpdateRipProgress(void);
char *FindRoot(char *str);
void MakeArgs(char *str,char **args,int maxargs);
void ParseID3Cmd(char *instr,char *outstr,int track, char *filename);
void ParseRipCmd(char *instr,char *outstr,int track,int startsec,int endsec,
		 char *filename);
void ParseMP3Cmd(char *instr,char *outstr,char *ripfile,char *mp3);
int MP3Encode(int track,int cpu);

void RedirectIO(gboolean redirect_output);
void ID3Add(char *file,int track);
void DoRip(GtkWidget *widget,gpointer data);
void MakeDirs(char *path);
char *MakePath(char *str);
gboolean RipNextTrack(void);
void PlaySegmentCB(GtkWidget *widget,gpointer data);
void PlaySegment(int track);
void ID3GenreChanged(GtkWidget *widget,gint number);
GtkWidget *MakeRangeSelects(void);
void MakeRipPage(void);
#endif


GdkColor gdkblack;
GdkColor gdkwhite;

GdkCursor *wait_cursor;

int cd_desc;
int changer_slots;
int current_disc=0;
gboolean local_mode=FALSE;
struct disc_info info;
DiscData ddata;
ProxyServer proxy_server={"",8000};
CDDBServer dbserver={"freedb.freedb.org","~cddb/cddb.cgi",
		     80,0,&proxy_server};
CDDBServer dbserver2={"","~cddb/cddb.cgi",
		      80,0,&proxy_server};
char cddevice[80]="/dev/cdrom";
int current_discid;
int first_time=1;
int root=0;
gboolean do_debug=FALSE;
gboolean stopped=FALSE;
gboolean playing=FALSE;
int looking_up=0;
gboolean ask_submit=FALSE;
int update_required=1;
int have_disc=0;
#ifndef GRIPCD
int minimized=0;
#else
int minimized=1;
#endif
gboolean volvis=FALSE;
gboolean track_edit_visible=FALSE;
gboolean track_prog_visible=FALSE;
int ffwding=0;
int rewinding=0;
int time_display_mode=TIME_MODE_TRACK;

/* The play mode */
unsigned long int play_mode = 0;
gboolean playloop=TRUE;

/* Related to program management */
int tracks_prog[MAX_TRACKS];
int prog_totaltracks    = 0;
int current_track_index = 0;
#define PREV_TRACK    (tracks_prog[current_track_index - 1])
#define CURRENT_TRACK (tracks_prog[current_track_index])
#define NEXT_TRACK    (tracks_prog[current_track_index + 1])

GtkWidget *window;
GtkWidget *winbox;
GtkWidget *controls;
GtkWidget *playopts;
GtkWidget *notebook;
GtkWidget *trackpage;
GtkWidget *helppage;
GtkWidget *configpage;
GtkWidget *config_notebook;
GtkWidget *aboutpage;
GtkWidget *trackclist;
#ifndef GRIPCD
gchar *titles[3]={"Track","Length","Rip"};
#else
gchar *titles[3]={"Track","Length","PL"};
#endif
GtkWidget *disc_name_label;
GtkWidget *disc_artist_label;
GtkWidget *current_track_label;
GtkWidget *play_time_label;
GtkWidget *play_sector_label;
GtkWidget *volume_control;
GtkWidget *track_edit_entry;
GtkWidget *title_edit_entry;
GtkWidget *artist_edit_entry;
GtkWidget *track_edit_box;
GtkWidget *playlist_entry;
GtkWidget *wht_check_image;
GtkWidget *blk_check_image;
GtkWidget *empty_image;
GtkWidget *cddb_indicator;
GtkWidget *play_indicator;
GtkWidget *cddb_pix[2];
GtkWidget *play_pix[3];

char cdupdate[256]="";

gboolean use_proxy=FALSE;
gboolean use_proxy_env=FALSE;

char ripexename[256]="/usr/bin/cdparanoia";
char ripcmdline[256]="-d %c %t:[.%b]-%t:[.%e] %f";
char ripfileformat[256]="%n.wav";
char mp3exename[256]="/usr/bin/bladeenc";
char mp3cmdline[256]="-QUIT %f";
char mp3fileformat[256]="%n.mp3";
char outputdir[256]="./%a/%d/";
int ripnice=0;
int mp3nice=0;
char underscore_chars[256]="\"/?'`/\\";
char id3exename[256]="/usr/bin/mp3info";
char id3cmdline[256]="-t %n -a %a -l %d -g %g %f";
gboolean rip_partial=FALSE;
gboolean delete_wavs=TRUE;
gboolean add_to_db=FALSE;
gboolean underscore_spaces=TRUE;
gboolean doid3=TRUE;
gboolean no_interrupt=FALSE;
gboolean stop_first=FALSE;
gboolean play_first=TRUE;
gboolean automatic_cddb=TRUE;
gboolean automatic_reshuffle=TRUE;
int start_sector,end_sector;
int max_wavs=99;
char cddb_submit_email[256]="freedb-submit@freedb.org";

GtkWidget *cddb_genre_combo;
GList *cddb_genre_item_list=NULL;
int num_cpu=1;
int edit_num_cpu=1;

/* Rip-related variables */

#ifndef GRIPCD
int doencode=0;
GtkWidget *rippage;
GtkWidget *ripprogbar;
GtkWidget *mp3progbar[MAX_NUM_CPU];

int stop_rip=0;
int ripping=0;
char ripfile[1024];
int ripsize;
int rippid;
int rip_quarter;
int rip_track;
char rip_delete_file[MAX_NUM_CPU][1024];
char mp3file[MAX_NUM_CPU][1024];
int mp3pid[MAX_NUM_CPU];
int encoding=0;
int mp3size[MAX_NUM_CPU];
int num_wavs;
int encode_track;
GtkWidget *riptracknamelabel;
GtkWidget *riptracktimelabel;
GtkWidget *rip_prog_label;
GtkWidget *mp3_prog_label[MAX_NUM_CPU];
GtkWidget *rip_indicator;
GtkWidget *rip_pix[4];
GtkWidget *mp3_indicator[MAX_NUM_CPU];
GtkWidget *mp3_pix[4];

GtkWidget *id3_genre_combo;
GList *id3_genre_item_list=NULL;
gint id3_genre_number = 12; /* default to "Other" */

/* this array contains string representations of all known ID3 tags */
/* taken from mp3id3 in the mp3tools 0.7 package */

gchar *id3_genres[] = {
  "Blues",
  "Classic Rock",
  "Country",
  "Dance",
  "Disco",
  "Funk",
  "Grunge",
  "Hip-Hop",
  "Jazz",
  "Metal",
  "New Age",
  "Oldies",
  "Other",
  "Pop",
  "R&B",
  "Rap",
  "Reggae",
  "Rock",
  "Techno",
  "Industrial",
  "Alternative",
  "Ska",
  "Death Metal",
  "Pranks",
  "Soundtrack",
  "Euro-Techno",
  "Ambient",
  "Trip-Hop",
  "Vocal",
  "Jazz+Funk",
  "Fusion",
  "Trance",
  "Classical",
  "Instrumental",
  "Acid",
  "House",
  "Game",
  "Sound Clip",
  "Gospel",
  "Noise",
  "AlternRock",
  "Bass",
  "Soul",
  "Punk",
  "Space",
  "Meditative",
  "Instrumental Pop",
  "Instrumental Rock",
  "Ethnic",
  "Gothic",
  "Darkwave",
  "Techno-Industrial",
  "Electronic",
  "Pop-Folk",
  "Eurodance",
  "Dream",
  "Southern Rock",
  "Comedy",
  "Cult",
  "Gangsta",
  "Top 40",
  "Christian Rap",
  "Pop/Funk",
  "Jungle",
  "Native American",
  "Cabaret",
  "New Wave",
  "Psychadelic",
  "Rave",
  "Showtunes",
  "Trailer",
  "Lo-Fi",
  "Tribal",
  "Acid Punk",
  "Acid Jazz",
  "Polka",
  "Retro",
  "Musical",
  "Rock & Roll",
  "Hard Rock",
  "Folk",
  "Folk/Rock",
  "National Folk",
  "Swing",
  "Bebob",
  "Latin",
  "Revival",
  "Celtic",
  "Bluegrass",
  "Avantgarde",
  "Gothic Rock",
  "Progressive Rock",
  "Psychedelic Rock",
  "Symphonic Rock",
  "Slow Rock",
  "Big Band",
  "Chorus",
  "Easy Listening",
  "Acoustic",
  "Humour",
  "Speech",
  "Chanson",
  "Opera",
  "Chamber Music",
  "Sonata",
  "Symphony",
  "Booty Bass",
  "Primus",
  "Porn Groove",
  "Satire",
  "Slow Jam",
  "Club",
  "Tango",
  "Samba",
  "Folklore",
  "Ballad",
  "Power Ballad",
  "Rhythmic Soul",
  "Freestyle",
  "Duet",
  "Punk Rock",
  "Drum Solo",
  "Acapella",
  "Euro-house",
  NULL
};

/* This array maps CDDB_ genre numbers to closest id3 genre */
int genre_cdda_2_id3[] =
{
  12,         /* CDDB_UNKNOWN   = 12 - other */
  0,          /* CDDB_BLUES     = 0 */
  32,         /* CDDB_CLASSICAL = 32 */
  2,          /* CDDB_COUNTRY   = 2 */
  12,         /* CDDB_DATA      = 12 - other */
  80,         /* CDDB_FOLK      = 80 */
  8,          /* CDDB_JAZZ      = 8 */
  12,         /* CDDB_MISC      = 12 - other */
  10,         /* CDDB_NEWAGE    = 10 */
  16,         /* CDDB_REGGAE    = 16 */
  17,         /* CDDB_ROCK      = 17 */
  24,         /* CDDB_SOUNDTRACK= 24 */
};
#endif

char *strvals[]={ripexename,ripcmdline,dbserver.name,
		 mp3exename,mp3cmdline,ripfileformat,mp3fileformat,
		 outputdir,proxy_server.name,cdupdate,underscore_chars,
		 id3exename,id3cmdline,dbserver.cgi_prog,
		 cddb_submit_email,dbserver2.name,dbserver2.cgi_prog};
GtkWidget *strentries[17];

int *intvals[]={&start_sector,&end_sector,&(proxy_server.port),
		&ripnice,&mp3nice,&max_wavs,&edit_num_cpu};
GtkWidget *intentries[7];

gboolean *checkvals[]={&rip_partial,&delete_wavs,&use_proxy,
		       &underscore_spaces,&doid3,&use_proxy_env,
                       &no_interrupt,&stop_first,&play_first,
                       &automatic_cddb,&automatic_reshuffle,
		       &add_to_db};
GtkWidget *checkbuttons[12];


CFGEntry cfg_entries[]={
  {"ripexename",CFG_ENTRY_STRING,256,ripexename},
  {"ripcmdline",CFG_ENTRY_STRING,256,ripcmdline},
  {"mp3exename",CFG_ENTRY_STRING,256,mp3exename},
  {"mp3cmdline",CFG_ENTRY_STRING,256,mp3cmdline},
  {"dbserver",CFG_ENTRY_STRING,256,dbserver.name},
  {"ripfileformat",CFG_ENTRY_STRING,256,ripfileformat},
  {"mp3fileformat",CFG_ENTRY_STRING,256,mp3fileformat},
  {"delete_wavs",CFG_ENTRY_BOOL,0,&delete_wavs},
  {"add_to_db",CFG_ENTRY_BOOL,0,&add_to_db},
  {"outputdir",CFG_ENTRY_STRING,256,outputdir},
  {"use_proxy",CFG_ENTRY_BOOL,0,&(dbserver.use_proxy)},
  {"proxy_name",CFG_ENTRY_STRING,256,proxy_server.name},
  {"proxy_port",CFG_ENTRY_INT,0,&(proxy_server.port)},
  {"cdupate",CFG_ENTRY_STRING,256,cdupdate},
  {"underscore_spaces",CFG_ENTRY_BOOL,0,&underscore_spaces},
  {"ripnice",CFG_ENTRY_INT,0,&ripnice},
  {"mp3nice",CFG_ENTRY_INT,0,&mp3nice},
  {"underscore_chars",CFG_ENTRY_STRING,256,underscore_chars},
  {"id3exename",CFG_ENTRY_STRING,256,id3exename},
  {"id3cmdline",CFG_ENTRY_STRING,256,id3cmdline},
  {"doid3",CFG_ENTRY_BOOL,0,&doid3},
  {"max_wavs",CFG_ENTRY_INT,0,&max_wavs},
  {"use_proxy_env",CFG_ENTRY_BOOL,0,&use_proxy_env},
  {"db_cgi",CFG_ENTRY_STRING,256,dbserver.cgi_prog},
  {"cddb_submit_email",CFG_ENTRY_STRING,256,cddb_submit_email},
  {"dbserver2",CFG_ENTRY_STRING,256,dbserver2.name},
  {"db2_cgi",CFG_ENTRY_STRING,256,dbserver2.cgi_prog},
  {"no_interrupt",CFG_ENTRY_BOOL,0,&no_interrupt},
  {"stop_first",CFG_ENTRY_BOOL,0,&stop_first},
  {"play_first",CFG_ENTRY_BOOL,0,&play_first},
  {"automatic_cddb",CFG_ENTRY_BOOL,0,&automatic_cddb},
  {"automatic_reshuffle",CFG_ENTRY_BOOL,0,&automatic_reshuffle},
  {"num_cpu",CFG_ENTRY_INT,0,&num_cpu},
  {"",CFG_ENTRY_LAST,0,NULL}
};

void ShutDown(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit();
}

gboolean FileExists(char *filename)
{
  struct stat mystat;

  return (stat(filename,&mystat)>=0);
}

void UpdateGTK(void)
{
  while(gtk_events_pending())
    gtk_main_iteration();
}

void Debug(char *fmt,...)
{
  va_list args;

  if(do_debug) {
    args=va_start(args,fmt);

    vprintf(fmt,args);
  }
}

void Busy(void)
{
  gdk_window_set_cursor(window->window,wait_cursor);

  UpdateGTK();
}

void UnBusy(void)
{
  gdk_window_set_cursor(window->window,NULL);

  UpdateGTK();
}

void DisplayMsg(GtkWidget *widget,GdkEventButton *event,gpointer data)
{
  BoolDialog((char *)data,"OK",NULL,NULL,NULL);
}

void BoolDialog(char *question,char *yes,GtkSignalFunc yesfunc,
		char *no,GtkSignalFunc nofunc)
{
  GtkWidget *dialog;
  GtkWidget *label;
  GtkWidget *yesbutton;
  GtkWidget *nobutton;

  dialog=gtk_dialog_new();

  gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),5);

  label=gtk_label_new(question);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),label,TRUE,TRUE,0);
  gtk_widget_show(label);

  yesbutton=gtk_button_new_with_label(yes);
  if(yesfunc)
    gtk_signal_connect(GTK_OBJECT(yesbutton),"clicked",
		       yesfunc,NULL);
  gtk_signal_connect_object(GTK_OBJECT(yesbutton),"clicked",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy),
			    GTK_OBJECT(dialog));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),yesbutton,
		     TRUE,TRUE,0);
  gtk_widget_show(yesbutton);

  if(no) {
    nobutton=gtk_button_new_with_label(no);
    if(nofunc)
      gtk_signal_connect(GTK_OBJECT(nobutton),"clicked",
			 nofunc,NULL);
    gtk_signal_connect_object(GTK_OBJECT(nobutton),"clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),nobutton,
		       TRUE,TRUE,0);
    gtk_widget_show(nobutton);
  }

  gtk_widget_show(dialog);

  gtk_grab_add(dialog);
}

void MinMax(void)
{
  if(minimized) {
    gtk_container_border_width(GTK_CONTAINER(winbox),3);
    gtk_widget_show(notebook);
  }
  else {
    gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,TRUE);
    gtk_container_border_width(GTK_CONTAINER(winbox),0);
    gtk_widget_hide(notebook);
  }

  minimized=!minimized;
}

void ToggleVol(void)
{
  if(volvis) {
    gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,TRUE);
    gtk_widget_hide(volume_control);
  }
  else {
    gtk_widget_show(volume_control);
  }

  volvis=!volvis;
}

void ToggleProg(void)
{
  if(track_prog_visible) {
    gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,TRUE);
    gtk_widget_hide(playopts);
  }
  else {
    gtk_widget_show(playopts);
  }

  track_prog_visible=!track_prog_visible;
}

void PlaylistChanged(void)
{
  strcpy(ddata.data_playlist,gtk_entry_get_text(GTK_ENTRY(playlist_entry)));

  InitProgram(info.disc_totaltracks,play_mode);

  BeUser();
  if(CDDBWriteDiscData(cd_desc,&ddata,NULL)<0)
    printf("Error saving disc data\n");
  BeRoot();
}

void ToggleTrackEdit(void)
{
  if(track_edit_visible) {
    gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,TRUE);
    gtk_widget_hide(track_edit_box);
  }
  else {
    gtk_widget_show(track_edit_box);
  }

  track_edit_visible=!track_edit_visible;
}

void SetTitle(char *title)
{
  gtk_entry_set_text(GTK_ENTRY(title_edit_entry),title);
}

void SetArtist(char *artist)
{
  gtk_entry_set_text(GTK_ENTRY(artist_edit_entry),artist);
}

void SetCDDBGenre(int genre)
{
  GtkWidget *item;

  item=GTK_WIDGET(g_list_nth(cddb_genre_item_list,genre)->data);
  gtk_list_select_child(GTK_LIST(GTK_COMBO(cddb_genre_combo)->list),
			item);  
}

void SaveDiscInfo(void)
{
  if(have_disc) {
    BeUser();
    if(CDDBWriteDiscData(cd_desc,&ddata,NULL)<0)
      printf("Error saving disc data\n");
    BeRoot();
  }
  else BoolDialog("No disc present","OK",NULL,NULL,NULL);
}

void TitleEditChanged(void)
{
  strcpy(ddata.data_title,gtk_entry_get_text(GTK_ENTRY(title_edit_entry)));
  gtk_label_set(GTK_LABEL(disc_name_label),ddata.data_title);
}

void ArtistEditChanged(void)
{
  strcpy(ddata.data_artist,gtk_entry_get_text(GTK_ENTRY(artist_edit_entry)));
  gtk_label_set(GTK_LABEL(disc_artist_label),ddata.data_artist);
}

void CDDBGenreChanged()
{
  ddata.data_genre=
    CDDBGenreValue(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cddb_genre_combo)->
						entry)));
}

void TrackEditChanged(void)
{
  char newname[1024];

  strcpy(ddata.data_track[CURRENT_TRACK].track_name,
	 gtk_entry_get_text(GTK_ENTRY(track_edit_entry)));

  snprintf(newname,1024,"%02d  %s",CURRENT_TRACK+1,
	   ddata.data_track[CURRENT_TRACK].track_name);
  
  gtk_clist_set_text(GTK_CLIST(trackclist),CURRENT_TRACK,0,newname);
}

gint TimeOut(gpointer data)
{
  /* Reset the update policy in case we changed it in MinMax() */

  gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,FALSE);

  if(ffwding) FastFwd();
  if(rewinding) Rewind();

#ifdef GRIPCD
  if(!have_disc)
    CheckNewDisc();

  UpdateDisplay();
#else
  if(ripping|encoding) UpdateRipProgress();
  else {
    if(!have_disc)
      CheckNewDisc();

    UpdateDisplay();
  }
#endif

  return TRUE;
}

char *UnderScore(char *str)
{
#ifndef GRIPCD
  char *c;

  for(c=str;*c;c++) {
    if(underscore_spaces&&(*c==' ')) *c='_';

    if(index(underscore_chars,*c)) *c='_';
  }
#endif

  return str;
}

void BeRoot(void)
{
  seteuid(0);
  setegid(0);
}

void BeUser(void)
{
  seteuid(getuid());
  setegid(getgid());
}

void SetVolume(GtkWidget *widget,gpointer data)
{
  struct disc_volume vol;

  vol.vol_front.left=vol.vol_front.right=
    vol.vol_back.left=vol.vol_back.right=GTK_ADJUSTMENT(widget)->value;

  CDSetVolume(cd_desc,&vol);
}

void NextTrack(void)
{
  CDStat(cd_desc,&info,FALSE);

  if(current_track_index<(prog_totaltracks-1)) {
    gtk_clist_select_row(GTK_CLIST(trackclist),
			 NEXT_TRACK,0);
  }
  else {
    if(playloop)
      gtk_clist_select_row(GTK_CLIST(trackclist),
			   tracks_prog[0],0);
    else stopped=TRUE;
  }
}

void PrevTrack(void)
{
  CDStat(cd_desc,&info,FALSE);

  if((info.disc_mode==CDAUDIO_PLAYING) &&
     ((info.disc_frame - info.track[info.disc_track-1].track_start) > 100))
    PlayTrack(CURRENT_TRACK);
  else {
    if(current_track_index)
      gtk_clist_select_row(GTK_CLIST(trackclist),
			   PREV_TRACK,0);
    else {
      if(playloop)
	gtk_clist_select_row(GTK_CLIST(trackclist),
			     tracks_prog[prog_totaltracks-1]
			     ,0);
    }
  }
}

void InitProgram(int num_tracks,int mode)
{
  int track;
  char *tok;

  if(mode==PM_PLAYLIST) {
    prog_totaltracks=0;

    tok=strtok(gtk_entry_get_text(GTK_ENTRY(playlist_entry)),",");

    while(tok) {
      tracks_prog[prog_totaltracks++]=atoi(tok)-1;

      tok=strtok(NULL,",");
    }
  }
  else {
    prog_totaltracks=num_tracks;
    
    for(track=0;track<num_tracks;track++) {
      tracks_prog[track]=track;
    }
    
    if(mode==PM_SHUFFLE)
      
      /* Shuffle the tracks around a bit */

      ShuffleTracks(num_tracks);
  }
}

void ShuffleTracks(int num_tracks)
{
  int t1,t2,tmp,shuffle;

  for(shuffle=0;shuffle<(num_tracks*10);shuffle++) {
    t1=RRand(num_tracks);
    t2=RRand(num_tracks);
    
    tmp=tracks_prog[t1];
    tracks_prog[t1]=tracks_prog[t2];
    tracks_prog[t2]=tmp;
  }
}

void ToggleLoop(void)
{
  playloop=!playloop;
}

void ChangePlayMode(void)
{
  play_mode=(play_mode+1)%PM_LASTMODE;
  CopyPixmap(GTK_PIXMAP(play_pix[play_mode]),GTK_PIXMAP(play_indicator));

  gtk_widget_set_sensitive(GTK_WIDGET(playlist_entry),play_mode==PM_PLAYLIST);

  InitProgram(info.disc_totaltracks,play_mode);
}

void FastFwdCB(GtkWidget *widget,gpointer data)
{
  ffwding=(int)data;

  if(ffwding) FastFwd();
}

void FastFwd(void)
{
  struct disc_timeval tv;

  tv.minutes=0;
  tv.seconds=5;

  if((info.disc_mode==CDAUDIO_PLAYING)||(info.disc_mode==CDAUDIO_PAUSED)) {
    CDAdvance(cd_desc,&info,&tv);
  }
}

void RewindCB(GtkWidget *widget,gpointer data)
{
  rewinding=(int)data;

  if(rewinding) Rewind();
}

void Rewind(void)
{
  struct disc_timeval tv;

  tv.minutes=0;
  tv.seconds=-5;

  if((info.disc_mode==CDAUDIO_PLAYING)||(info.disc_mode==CDAUDIO_PAUSED)) {
    CDAdvance(cd_desc,&info,&tv);
  }
}

void NextDisc(void)
{
  if(changer_slots>1) {
    current_disc=(current_disc+1)%changer_slots;
    CDChangerSelectDisc(cd_desc,current_disc);
    have_disc=0;
  }
}

void StopPlay(void)
{
  CDStop(cd_desc);
  CDStat(cd_desc,&info,FALSE);
  stopped=TRUE;

  if (stop_first)
    gtk_clist_select_row(GTK_CLIST(trackclist),0,0);
}

void EjectDisc(void)
{
  Busy();

  if(have_disc) {
    CDStop(cd_desc);
    CDEject(cd_desc);
    playing=0;
    have_disc=0;
    update_required=1;
    current_discid=0;
  }
  else CDClose(cd_desc);

  UnBusy();
}

void PlayTrackCB(GtkWidget *widget,gpointer data)
{
  int track;

  CDStat(cd_desc,&info,FALSE);

  if(play_mode!=PM_NORMAL&&!((info.disc_mode==CDAUDIO_PLAYING)||
			     info.disc_mode==CDAUDIO_PAUSED)) {
    if(play_mode == PM_SHUFFLE && automatic_reshuffle)
      ShuffleTracks(info.disc_totaltracks);
    current_track_index=0;
    gtk_clist_select_row(GTK_CLIST(trackclist),CURRENT_TRACK,0);
  }

  track=CURRENT_TRACK;

  if(track==(info.disc_track-1)) {
    switch(info.disc_mode) {
    case CDAUDIO_PLAYING:
      CDPause(cd_desc);
      return;
      break;
    case CDAUDIO_PAUSED:
      CDResume(cd_desc);
      return;
      break;
    default:
      PlayTrack(track);
      break;
    }
  }
  else PlayTrack(track);
}

void PlayTrack(int track)
{
  Busy();
  
  if(play_mode==PM_NORMAL)
    CDPlayTrack(cd_desc,&info,track+1,info.disc_totaltracks);
  else CDPlayTrack(cd_desc,&info,track+1,track+1);

  UnBusy();

  playing=TRUE;
}

void ChangeTimeMode(void)
{
  time_display_mode=(time_display_mode+1)%4;
  UpdateDisplay();
}

void CheckNewDisc(void)
{
  int new_id;

  if(!looking_up) {
    Debug("Checking for a new disc\n");

    CDStat(cd_desc,&info,FALSE);

    if(info.disc_present) {
      CDStat(cd_desc,&info,TRUE);

      Debug("CDStat found a disc, checking tracks\n");
      
      if(CheckTracks()) {
	Debug("We have a valid disc!\n");
	
	new_id=CDDBDiscid(cd_desc);
	InitProgram(info.disc_totaltracks,play_mode);
        if(play_first)
          if(info.disc_mode == CDAUDIO_COMPLETED ||
	     info.disc_mode == CDAUDIO_NOSTATUS) {
            gtk_clist_select_row(GTK_CLIST(trackclist),0,0);
            info.disc_track = 1;
          }
	
	if(new_id) {
	  have_disc=1;
	  LookupDisc(FALSE);
	}
      }
      else Debug("No non-zero length tracks\n");
    }
    else Debug("CDStat said no disc\n");
  }
}

/* Check to make sure we didn't get a false alarm from the cdrom device */

gboolean CheckTracks(void)
{
  int track;
  gboolean have_track=FALSE;

  for(track=0;track<info.disc_totaltracks;track++)
    if(info.track[track].track_length.minutes||
       info.track[track].track_length.seconds) have_track=TRUE;

  return have_track;
}

void UpdateDisplay(void)
{
  static int play_counter=0;
  static int cddb_counter=0;
  char buf[80]="";
  char icon_buf[80];
  static int frames;
  static int secs;
  static int mins;
  int totsecs;

  if(!looking_up) {
    if(cddb_counter%2)
      cddb_counter++;
  }
  else
    CopyPixmap(GTK_PIXMAP(cddb_pix[cddb_counter++%2]),
	       GTK_PIXMAP(cddb_indicator));

  if(!update_required) {
    if(have_disc) {
      CDStat(cd_desc,&info,FALSE);

      if((info.disc_mode==CDAUDIO_PLAYING)||
	 (info.disc_mode==CDAUDIO_PAUSED)) {
	if(info.disc_mode==CDAUDIO_PAUSED) {
	  if((play_counter++%2)==0) {
	    strcpy(buf,"");
	  }
	  else {
	    snprintf(buf,80,"%02d:%02d",mins,secs);
	  }
	}
	else {
	  if((info.disc_track-1)!=CURRENT_TRACK)
	    gtk_clist_select_row(GTK_CLIST(trackclist),info.disc_track-1,0);

	  frames=info.disc_frame-info.track[info.disc_track-1].track_start;

	  switch(time_display_mode) {
	  case TIME_MODE_TRACK:
	    mins=info.track_time.minutes;
	    secs=info.track_time.seconds;
	    break;
	  case TIME_MODE_DISC:
	    mins=info.disc_time.minutes;
	    secs=info.disc_time.seconds;
	    break;
	  case TIME_MODE_LEFT_TRACK:
	    secs=(info.track_time.minutes*60)+info.track_time.seconds;
	    totsecs=(info.track[CURRENT_TRACK].track_length.minutes*60)+
	      info.track[CURRENT_TRACK].track_length.seconds;
	    
	    totsecs-=secs;
	    
	    mins=totsecs/60;
	    secs=totsecs%60;
	    break;
	  case TIME_MODE_LEFT_DISC:
	    secs=(info.disc_time.minutes*60)+info.disc_time.seconds;
	    totsecs=(info.disc_length.minutes*60)+info.disc_length.seconds;
	    
	    totsecs-=secs;
	    
	    mins=totsecs/60;
	    secs=totsecs%60;
	    break;
	  }
	  
#ifndef GRIPCD
	  snprintf(buf,80,"Current sector: %6d",frames);
	  gtk_label_set(GTK_LABEL(play_sector_label),buf);
#endif
          if (time_display_mode == TIME_MODE_LEFT_TRACK ||
              time_display_mode == TIME_MODE_LEFT_DISC)
            snprintf(buf,80,"-%02d:%02d",mins,secs);
          else
	    snprintf(buf,80,"%02d:%02d",mins,secs);
	}
      }
      else {
	if(playing&&((info.disc_mode==CDAUDIO_COMPLETED)||
	   ((info.disc_mode==CDAUDIO_NOSTATUS)&&!stopped))) {
	  NextTrack();
	  strcpy(buf,"00:00");
	  if(!stopped) PlayTrack(CURRENT_TRACK);
	}
	else if(stopped) {
	  CDStop(cd_desc);
#ifndef GRIPCD
	  frames=secs=mins=0;
	  snprintf(buf,80,"Current sector: %6d",frames);
	  gtk_label_set(GTK_LABEL(play_sector_label),buf);
#endif
	  
	  strcpy(buf,"00:00");
	  
	  stopped=FALSE;
	  playing=FALSE;
	}
	else return;
      }
      
      gtk_label_set(GTK_LABEL(play_time_label),buf);
      snprintf(icon_buf, sizeof(icon_buf), "%02d %s %s",
	       info.disc_track, buf, PROGRAM);
      gdk_window_set_icon_name(window->window, icon_buf);
    }
  }

  if(update_required) {
    UpdateTracks();
    
    if(have_disc) {
      snprintf(buf,80,"%02d:%02d",info.disc_length.minutes,
	       info.disc_length.seconds);
      snprintf(icon_buf, sizeof(icon_buf), "%02d %s %s",
	       info.disc_track, buf, PROGRAM);
	       
      gtk_label_set(GTK_LABEL(play_time_label),buf);
      
      if(!looking_up)
	CopyPixmap(GTK_PIXMAP(empty_image),GTK_PIXMAP(cddb_indicator));
      
      if(!no_interrupt)
        gtk_clist_select_row(GTK_CLIST(trackclist),0,0);
      else
        gtk_clist_select_row(GTK_CLIST(trackclist),info.disc_track-1,0);
    }
    else {
      gtk_label_set(GTK_LABEL(play_time_label),"--:--");
      strncpy(icon_buf, PROGRAM, sizeof(icon_buf));
      
      SetCurrentTrack(-1);
    }
    gdk_window_set_icon_name(window->window, icon_buf);
  }
}

void setCurrentTrackIndex(int track)
{
  /* Looks up the track of index track in the program */
  for(current_track_index = 0;
      (current_track_index < MAX_TRACKS)
	&& (current_track_index < prog_totaltracks)
	&& (CURRENT_TRACK != track);
      current_track_index++)
    continue;
}

void SetCurrentTrack(int track)
{
  char buf[256];
#ifndef GRIPCD
  int tracklen;
#endif

  if(track<0) {
    gtk_label_set(GTK_LABEL(current_track_label),"--");
#ifndef GRIPCD
    gtk_entry_set_text(GTK_ENTRY(intentries[0]),"0");
    gtk_entry_set_text(GTK_ENTRY(intentries[1]),"0");
    gtk_label_set(GTK_LABEL(riptracknamelabel),"Track: n/a");
    gtk_label_set(GTK_LABEL(riptracktimelabel),"");
#endif
  }
  else {
    gtk_signal_handler_block_by_func(GTK_OBJECT(track_edit_entry),
				     TrackEditChanged,NULL);
    gtk_entry_set_text(GTK_ENTRY(track_edit_entry),
		       ddata.data_track[track].track_name);

    gtk_signal_handler_unblock_by_func(GTK_OBJECT(track_edit_entry),
				       TrackEditChanged,NULL);
    snprintf(buf,80,"%02d",track+1);
    gtk_label_set(GTK_LABEL(current_track_label),buf);
	
#ifndef GRIPCD
    gtk_entry_set_text(GTK_ENTRY(intentries[0]),"0");
	
    tracklen=(info.track[track+1].track_start-1)-info.track[track].track_start;
    snprintf(buf,80,"%d",tracklen);
    gtk_entry_set_text(GTK_ENTRY(intentries[1]),buf);
	
    snprintf(buf,80,"Track: %s",ddata.data_track[track].track_name);
    gtk_label_set(GTK_LABEL(riptracknamelabel),buf);

    snprintf(buf,80,"Playing time: %d:%02d",
	     info.track[track].track_length.minutes,
	    info.track[track].track_length.seconds);
    gtk_label_set(GTK_LABEL(riptracktimelabel),buf);
#endif

    setCurrentTrackIndex(track);
  }
}

#ifndef GRIPCD
gboolean TrackIsChecked(int track)
{
  GdkPixmap *pm=NULL;
  GdkBitmap *bm=NULL;

  gtk_clist_get_pixmap(GTK_CLIST(trackclist),track,2,&pm,&bm);
  return((pm==GTK_PIXMAP(wht_check_image)->pixmap) ||
	 (pm==GTK_PIXMAP(blk_check_image)->pixmap));
}

void ToggleChecked(int track)
{
  SetChecked(track,!TrackIsChecked(track));
}

void SetChecked(int track,gboolean checked)
{
  if(!checked)
    gtk_clist_set_pixmap(GTK_CLIST(trackclist),track,2,
			 GTK_PIXMAP(empty_image)->pixmap,
			 GTK_PIXMAP(empty_image)->mask);
  else {
    if(track==CURRENT_TRACK)
      gtk_clist_set_pixmap(GTK_CLIST(trackclist),track,2,
			   GTK_PIXMAP(wht_check_image)->pixmap,
			   GTK_PIXMAP(wht_check_image)->mask);
    else
      gtk_clist_set_pixmap(GTK_CLIST(trackclist),track,2,
			   GTK_PIXMAP(blk_check_image)->pixmap,
			   GTK_PIXMAP(blk_check_image)->mask);
  }
}

void ClickColumn(GtkWidget *widget,gint column,gpointer data)
{
  int track;
  int numsel=0;

  for(track=0;track<info.disc_totaltracks;track++)
    if(TrackIsChecked(track)) numsel++;

  for(track=0;track<info.disc_totaltracks;track++)
    SetChecked(track,(numsel<info.disc_totaltracks/2));
}
#endif

void CListButtonPressed(GtkWidget *widget,GdkEventButton *event,gpointer data)
{
  gint row,col;

  if(event&&event->button!=1) {
    gtk_clist_get_selection_info(GTK_CLIST(trackclist),event->x,event->y,
				 &row,&col);

#ifndef GRIPCD
    ToggleChecked(row);
#endif
  }
}

void UnSelectRow(GtkWidget *widget,gint row,gint column,
		 GdkEventButton *event,gpointer data)
{
#ifndef GRIPCD
  if(TrackIsChecked(row))
    gtk_clist_set_pixmap(GTK_CLIST(trackclist),row,2,
			 GTK_PIXMAP(blk_check_image)->pixmap,
			 GTK_PIXMAP(blk_check_image)->mask);
#endif
}

void SelectRow(GtkWidget *widget,gint row,gint column,
	       GdkEventButton *event,gpointer data)
{
  SetCurrentTrack(row);

#ifndef GRIPCD
  if(TrackIsChecked(row))
    gtk_clist_set_pixmap(GTK_CLIST(trackclist),row,2,
			 GTK_PIXMAP(wht_check_image)->pixmap,
			 GTK_PIXMAP(wht_check_image)->mask);
#endif

  if(gtk_clist_row_is_visible(GTK_CLIST(trackclist),row)!=GTK_VISIBILITY_FULL)
    gtk_clist_moveto(GTK_CLIST(trackclist),row,0,0,0);

  if((info.disc_mode==CDAUDIO_PLAYING)&&(info.disc_track!=(row+1)))
    PlayTrack(row);
  else {
    if(event) {
      switch(event->type) {
      case GDK_2BUTTON_PRESS:
	PlayTrack(row);
	break;
      default:
	break;
      }
    }
  }
}

pthread_t cddb_thread;

void CDDBToggle(void)
{
  if(looking_up) {
    pthread_kill_other_threads_np();
    printf("Aborted\n");
    looking_up=0;
    update_required=1;
  }
  else
    if(have_disc) LookupDisc(TRUE);
}

void LookupDisc(gboolean manual)
{
  int track;
  gboolean present;

  present=CDDBStatDiscData(cd_desc);

  if(!manual&&present) {
    CDDBReadDiscData(cd_desc,&ddata);
    update_required=1;
  }
  else {
    if(!manual) {
      ddata.data_id=CDDBDiscid(cd_desc);
      ddata.data_genre=7; /* "misc" */
      strcpy(ddata.data_title,"Unknown Disc");
      strcpy(ddata.data_artist,"");
      
      for(track=0;track<info.disc_totaltracks;track++) {
	sprintf(ddata.data_track[track].track_name,"Track %d",track+1);
	*(ddata.data_track[track].track_extended)='\0';
      }

      *ddata.data_extended='\0';
      
      update_required=1;
    }

    if(!local_mode && manual?TRUE:automatic_cddb) {
      looking_up=1;
      
      pthread_create(&cddb_thread,NULL,(void *)&DoLookup,NULL);
    }
  }
}

void DoLookup(void)
{
  BeUser();

  if(!CDDBLookupDisc(&dbserver)) {
    if(*(dbserver2.name)) {
      if(CDDBLookupDisc(&dbserver2))
	ask_submit=TRUE;
    }
  }

  BeRoot();

  looking_up=0;
  pthread_exit(0);
}

gboolean CDDBLookupDisc(CDDBServer *server)
{
  CDDBHello hello;
  CDDBQuery query;
  CDDBEntry entry;
  gboolean success=FALSE;

  if(server->use_proxy)
    printf("Querying %s (through %s) for disc %02lx.\n",server->name,
	   server->proxy->name,
	   CDDBDiscid(cd_desc));
  else
    printf("Querying %s for disc %02lx.\n",server->name,
	   CDDBDiscid(cd_desc));

  strncpy(hello.hello_program,PROGRAM,256);
  strncpy(hello.hello_version,VERSION,256);
	
  if(!CDDBDoQuery(cd_desc,server,&hello,&query)) {
    printf("Query failed\n");
    update_required=1;
  } else {
    switch(query.query_match) {
    case MATCH_INEXACT:
    case MATCH_EXACT:
      printf("Match for \"%s / %s\"\nDownloading data...\n",
	     query.query_list[0].list_artist,
	     query.query_list[0].list_title);
      entry.entry_genre = query.query_list[0].list_genre;
      entry.entry_id = query.query_list[0].list_id;
      CDDBRead(cd_desc,server,&hello,&entry,&ddata);
		
      printf("Done\n");
      success=TRUE;
		
      if(CDDBWriteDiscData(cd_desc,&ddata,NULL)<0)
	printf("Error saving disc data\n");
		
      update_required=1;
      break;
    case MATCH_NOMATCH:
      printf("No match\n");
      break;
    }
  }

  return success;
}

void SubmitEntryCB(void)
{
  int len;

  if(!have_disc)
    return;

  len=strlen(cddb_submit_email);

  if(!strncasecmp(cddb_submit_email+(len-9),".cddb.com",9))
    BoolDialog("You are about to submit this disc information\n"
	       "to a commercial CDDB server, which will then\n"
	       "own the data that you submit. These servers make\n"
	       "a profit out of your effort. We suggest that you\n"
	       "support free CDDB servers instead.\n\nContinue?",
	       "Yes",SubmitEntry,"No",NULL);
  else
    BoolDialog("You are about to submit this\ndisc information via email.\n\n"
	       "Continue?",
	       "Yes",SubmitEntry,"No",NULL);
}

void SubmitEntry(void)
{
  FILE *efp;
  char *filename;
  char mailcmd[256];

  filename=tmpnam(NULL);

  efp=fopen(filename,"w");

  if(!efp) {
    printf("Error: Unable to create temporary file\n\n");
  }
  else {
    if(CDDBWriteDiscData(cd_desc,&ddata,efp)<0) {
      printf("Error: Unable to write disc data\n\n");
      fclose(efp);
    }
    else {
      fclose(efp);

      snprintf(mailcmd,256,"%s <%s -s \"cddb %s %02lx\" %s",
	       MAILER,filename,CDDBGenre(ddata.data_genre),ddata.data_id,
	       cddb_submit_email);

      printf("Mailing entry to %s\n",cddb_submit_email);

      BeUser();
      system(mailcmd);
      BeRoot();

      remove(filename);
    }
  }
}

#ifndef GRIPCD

void CheckDupNames(void)
{
  int track,track2;
  int numdups[MAX_TRACKS];
  int count;
  char buf[256];

  for(track=0;track<info.disc_totaltracks;track++)
    numdups[track]=0;
  
  for(track=0;track<(info.disc_totaltracks-1);track++) {
    if(!numdups[track]) {
      count=0;

      for(track2=track+1;track2<info.disc_totaltracks;track2++) {
	if(!strcmp(ddata.data_track[track].track_name,
		   ddata.data_track[track2].track_name))
	  numdups[track2]=++count;
      }
    }
  }

  for(track=0;track<info.disc_totaltracks;track++) {
    if(numdups[track]) {
      snprintf(buf,260,"%s (%d)",ddata.data_track[track].track_name,
	       numdups[track]+1);

      strcpy(ddata.data_track[track].track_name,buf);
    }
  }
}

#endif /* ifndef GRIPCD */

void UpdateTracks(void)
{
  int track;
  char buf[1024];
  char *col_strings[3];
#ifndef GRIPCD
  int id3_number;
  GtkWidget *id3_genre;
#endif

  if(have_disc) {
    current_discid=CDDBDiscid(cd_desc);

    SetTitle(ddata.data_title);
    SetArtist(ddata.data_artist);
    SetCDDBGenre(ddata.data_genre);

#ifndef GRIPCD
    id3_number=genre_cdda_2_id3[ddata.data_genre];
    id3_genre=GTK_WIDGET(g_list_nth(id3_genre_item_list,id3_number)->data);
    gtk_list_select_child(GTK_LIST(GTK_COMBO(id3_genre_combo)->list),
			  id3_genre);  
#endif
  }
  else {
    SetTitle("No Disc");
    SetArtist("");
    SetCDDBGenre(0);
  }

  gtk_entry_set_text(GTK_ENTRY(playlist_entry),ddata.data_playlist);

  if(*cdupdate) {
    ParseFileFmt(cdupdate,buf,0,0,0);
    BeUser();

    system(buf);
    BeRoot();
  }

  if(!first_time)
      gtk_clist_clear(GTK_CLIST(trackclist));
  else {
    setCurrentTrackIndex(info.disc_track - 1);
  }

  if(have_disc) {
#ifndef GRIPCD
    CheckDupNames();
#endif

    col_strings[0]=(char *)malloc(260);
    col_strings[1]=(char *)malloc(6);
    col_strings[2]=NULL;

    for(track=0;track<info.disc_totaltracks;track++) {
      snprintf(col_strings[0],260,"%02d  %s",track+1,
	       ddata.data_track[track].track_name);

      snprintf(col_strings[1],6,"%2d:%02d",
	       info.track[track].track_length.minutes,
	       info.track[track].track_length.seconds);

      gtk_clist_append(GTK_CLIST(trackclist),col_strings);
    }

    free(col_strings[0]);
    free(col_strings[1]);
  }

  if(ask_submit) {
    BoolDialog("This disc has been found on your secondary server,\n"
	       "but not on your primary server.\n\n"
	       "Do you wish to submit this disc information?",
	       "Yes",SubmitEntry,"No",NULL);

    ask_submit=FALSE;
  }

  first_time=0;
  update_required=0;
}

void ChangeStrVal(GtkWidget *widget,gpointer data)
{
  int num;;

  num=(int)data;

  strcpy(strvals[num],gtk_entry_get_text(GTK_ENTRY(widget)));

  if(num==7) chdir(outputdir);
}

GtkWidget *MakeStrEntry(char *name,int num,int len,gboolean editable)
{
  GtkWidget *label;
  GtkWidget *hbox;

  hbox=gtk_hbox_new(FALSE,5);

  label=gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  strentries[num]=gtk_entry_new_with_max_length(len);
  gtk_entry_set_editable(GTK_ENTRY(strentries[num]),editable);
  gtk_entry_set_text(GTK_ENTRY(strentries[num]),strvals[num]);
  gtk_signal_connect(GTK_OBJECT(strentries[num]),"changed",
		     GTK_SIGNAL_FUNC(ChangeStrVal),(gpointer)num);

  if(!editable)
    gtk_signal_connect(GTK_OBJECT(strentries[num]),"button_press_event",
		       GTK_SIGNAL_FUNC(DisplayMsg),
		       (gpointer)"Only Root can edit this field");

  gtk_box_pack_start(GTK_BOX(hbox),strentries[num],TRUE,TRUE,0);

  gtk_entry_set_position(GTK_ENTRY(strentries[num]),0);

  gtk_widget_show(strentries[num]);

  return hbox;
}

void ChangeIntVal(GtkWidget *widget,gpointer data)
{
  int num;

  num=(int)data;

  *(intvals[num])=atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
}

GtkWidget *MakeNumEntry(char *name,int num,int len)
{
  char buf[80];
  GtkWidget *label;
  GtkWidget *hbox;

  hbox=gtk_hbox_new(FALSE,5);

  label=gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  intentries[num]=gtk_entry_new_with_max_length(len);
  gtk_widget_set_usize(intentries[num],len*8+5,0);
  sprintf(buf,"%d",*(intvals[num]));
  gtk_entry_set_text(GTK_ENTRY(intentries[num]),buf);
  gtk_signal_connect(GTK_OBJECT(intentries[num]),"changed",
		     GTK_SIGNAL_FUNC(ChangeIntVal),(gpointer)num);

  gtk_box_pack_end(GTK_BOX(hbox),intentries[num],FALSE,FALSE,0);
  gtk_widget_show(intentries[num]);

  return hbox;
}

void ChangeBoolVal(GtkWidget *widget,gpointer data)
{
  int num;

  num=(int)data;

  *checkvals[num]=!*checkvals[num];

  if(num==2) dbserver.use_proxy=*checkvals[num];
}

GtkWidget *MakeCheckButton(char *name,int num)
{
  checkbuttons[num]=gtk_check_button_new_with_label(name);
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(checkbuttons[num]),
			      *(checkvals[num]));
  gtk_signal_connect(GTK_OBJECT(checkbuttons[num]),"clicked",
		     GTK_SIGNAL_FUNC(ChangeBoolVal),
		     (gpointer)num);

  return checkbuttons[num];
}

GtkWidget *MakeNewPage(char *name)
{
  GtkWidget *page;
  GtkWidget *label;

  page=gtk_frame_new(NULL);
  gtk_widget_show(page);

  label=gtk_label_new(name);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook),page,label);

  return page;
}

void MakeTrackPage(void)
{
  GtkWidget *vbox;
#if (GTK_MINOR_VERSION != 0)
  GtkWidget *scroll;
#endif

  trackpage=MakeNewPage("Tracks");

  vbox=gtk_vbox_new(FALSE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  disc_name_label=gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(vbox),disc_name_label,FALSE,FALSE,0);
  gtk_widget_show(disc_name_label);

  disc_artist_label=gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(vbox),disc_artist_label,FALSE,FALSE,0);
  gtk_widget_show(disc_artist_label);

#ifndef GRIPCD
  trackclist=gtk_clist_new_with_titles(3,titles);
#else
  trackclist=gtk_clist_new_with_titles(2,titles);
#endif

  gtk_clist_set_column_justification(GTK_CLIST(trackclist),1,
				     GTK_JUSTIFY_RIGHT);

  gtk_clist_set_selection_mode(GTK_CLIST(trackclist),GTK_SELECTION_SINGLE);
  gtk_clist_column_title_passive(GTK_CLIST(trackclist),0);
  gtk_clist_column_title_passive(GTK_CLIST(trackclist),1);
#ifndef GRIPCD
  gtk_clist_column_title_active(GTK_CLIST(trackclist),2);
  gtk_clist_set_column_width(GTK_CLIST(trackclist),0,WINWIDTH-120);
  gtk_clist_set_column_width(GTK_CLIST(trackclist),1,40);
  gtk_clist_set_column_justification(GTK_CLIST(trackclist),2,
				     GTK_JUSTIFY_CENTER);
#else
  gtk_clist_set_column_width(GTK_CLIST(trackclist),0,WINWIDTH-90);
#endif

  gtk_signal_connect(GTK_OBJECT(trackclist),"select_row",
		     GTK_SIGNAL_FUNC(SelectRow),
		     NULL);
  gtk_signal_connect(GTK_OBJECT(trackclist),"unselect_row",
		     GTK_SIGNAL_FUNC(UnSelectRow),
		     (gpointer)1);
  
  gtk_signal_connect(GTK_OBJECT(trackclist),"button_press_event",
		     GTK_SIGNAL_FUNC(CListButtonPressed),NULL);

#ifndef GRIPCD
  gtk_signal_connect(GTK_OBJECT(trackclist),"click_column",
		     GTK_SIGNAL_FUNC(ClickColumn),NULL);
#endif

#if (GTK_MINOR_VERSION == 0)
  gtk_clist_set_policy(GTK_CLIST(trackclist),GTK_POLICY_AUTOMATIC,
		       GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX(vbox),trackclist,TRUE,TRUE,0);
#else
  scroll=gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				 GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(scroll),trackclist);
  gtk_box_pack_start(GTK_BOX(vbox),scroll,TRUE,TRUE,0);
  gtk_widget_show(scroll);
#endif

  gtk_widget_show(trackclist);

  gtk_container_add(GTK_CONTAINER(trackpage),vbox);
  gtk_widget_show(vbox);
}

void ChopCR(char *str)
{
  int len;

  len=strlen(str);

  if(str[len-1]=='\n') str[len-1]='\0';
}

void DoLoadConfig(void)
{
  char filename[256];
  char *proxy_env,*tok;

  sprintf(filename,"%s/grip.cfg",AUXDIR);

  if(!LoadConfig(filename,"GRIP",1,1,cfg_entries))
    printf("Error: Unable to load config file from %s\n",filename);

  dbserver2.use_proxy=dbserver.use_proxy;

  if(use_proxy_env) {                /* Get proxy info from "http_proxy" */
    proxy_env=getenv("http_proxy");
    if(proxy_env) {
      
      /* Skip the "http://" if it's present */
      
      if(!strncasecmp(proxy_env,"http://",7)) proxy_env+=7;
      
      tok=strtok(proxy_env,":");
      if(tok) strncpy(proxy_server.name,tok,256);
      
      tok=strtok(NULL,"/");
      if(tok) proxy_server.port=atoi(tok);
      
      printf("server is %s, port %d\n",proxy_server.name,
	     proxy_server.port);
    }
  }
}

void DoSaveConfig(void)
{
  char filename[256];

  snprintf(filename,256,"%s/grip.cfg",AUXDIR);

  if(!SaveConfig(filename,"GRIP",1,cfg_entries))
    printf("Error: Unable to save config file to %s\n",filename);
}

void MakeConfigPage(void)
{
  GtkWidget *vbox,*vbox2,*dbvbox;
  GtkWidget *entry;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *page,*dbpage;
  GtkWidget *check;
  GtkWidget *notebook;

  configpage=MakeNewPage("Config");

  vbox2=gtk_vbox_new(FALSE,2);
  gtk_container_border_width(GTK_CONTAINER(vbox2),3);

  config_notebook=gtk_notebook_new();

  page=gtk_frame_new(NULL);
  vbox=gtk_vbox_new(TRUE,0);

  check=MakeCheckButton("Don't interrupt playback on exit/startup", 6);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  check=MakeCheckButton("Rewind when stopped", 7);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  check=MakeCheckButton("Startup with first track if not playing",8);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  check=MakeCheckButton("Reshuffle before each playback",10);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("CD");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);

#ifndef GRIPCD
  page=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  entry=MakeStrEntry("Ripping executable",0,255,root);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("Rip command-line",1,255,root);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("Rip file format",5,255,root);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeNumEntry("Rip 'nice' value",3,3);
  gtk_box_pack_start(GTK_BOX(vbox),entry,FALSE,FALSE,0);
  gtk_widget_show(entry);
  
  entry=MakeNumEntry("Max non-encoded .wav's",5,3);
  gtk_box_pack_start(GTK_BOX(vbox),entry,FALSE,FALSE,0);
  gtk_widget_show(entry);
  
  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Rip");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);

  page=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  entry=MakeStrEntry("MP3 executable",3,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("MP3 command-line",4,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("MP3 file format",6,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  check=MakeCheckButton("Delete .wav after encoding",1);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

#ifdef MP3DB
  check=MakeCheckButton("Insert info into SQL database",11);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);
#endif

  entry=MakeNumEntry("Number of CPUs to use",6,3);
  gtk_box_pack_start(GTK_BOX(vbox),entry,FALSE,FALSE,0);
  gtk_widget_show(entry);
  
  entry=MakeNumEntry("Mp3 'nice' value",4,3);
  gtk_box_pack_start(GTK_BOX(vbox),entry,FALSE,FALSE,0);
  gtk_widget_show(entry);
  
  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("MP3");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);

  page=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  entry=MakeStrEntry("ID3 executable",11,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  check=MakeCheckButton("Add ID3 tags to MP3 files",4);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  entry=MakeStrEntry("ID3 command-line",12,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);
  
  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label = gtk_label_new("ID3");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);
#endif

  page=gtk_frame_new(NULL);

  dbvbox=gtk_vbox_new(FALSE,2);
  gtk_container_border_width(GTK_CONTAINER(dbvbox),3);

  notebook=gtk_notebook_new();

  dbpage=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  entry=MakeStrEntry("DB server",2,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("CGI path",13,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  gtk_container_add(GTK_CONTAINER(dbpage),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Primary Server");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook),dbpage,label);
  gtk_widget_show(dbpage);


  dbpage=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  entry=MakeStrEntry("DB server",15,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeStrEntry("CGI path",16,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  gtk_container_add(GTK_CONTAINER(dbpage),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Secondary Server");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook),dbpage,label);
  gtk_widget_show(dbpage);


  gtk_box_pack_start(GTK_BOX(dbvbox),notebook,TRUE,TRUE,0);
  gtk_widget_show(notebook);


  entry=MakeStrEntry("DB Submit email",14,255,TRUE);
  gtk_box_pack_start(GTK_BOX(dbvbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  check=MakeCheckButton("Perform CDDB lookup automatically",9);
  gtk_box_pack_start(GTK_BOX(dbvbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  gtk_container_add(GTK_CONTAINER(page),dbvbox);
  gtk_widget_show(dbvbox);


  label=gtk_label_new("CDDB");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);


  page=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  check=MakeCheckButton("Use proxy server",2);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  check=MakeCheckButton("Get server from 'http_proxy' env. var",5);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  entry=MakeStrEntry("Proxy server",8,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  entry=MakeNumEntry("Proxy port",2,5);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Proxy");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);

  page=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

#ifndef GRIPCD
  entry=MakeStrEntry("Output directory",7,255,root);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  check=MakeCheckButton("Convert spaces to underscores",3);
  gtk_box_pack_start(GTK_BOX(vbox),check,TRUE,TRUE,0);
  gtk_widget_show(check);

  entry=MakeStrEntry("Also underscore",10,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);
#endif

  entry=MakeStrEntry("CD update program",9,255,TRUE);
  gtk_box_pack_start(GTK_BOX(vbox),entry,TRUE,TRUE,0);
  gtk_widget_show(entry);

  gtk_container_add(GTK_CONTAINER(page),vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Misc");
  gtk_notebook_append_page(GTK_NOTEBOOK(config_notebook),page,label);
  gtk_widget_show(page);

  gtk_box_pack_start(GTK_BOX(vbox2),config_notebook,TRUE,TRUE,0);
  gtk_widget_show(config_notebook);

  button=gtk_button_new_with_label("Save Configuration");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(DoSaveConfig),NULL);
  gtk_box_pack_start(GTK_BOX(vbox2),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  gtk_container_add(GTK_CONTAINER(configpage),vbox2);
  gtk_widget_show(vbox2);
}

void Homepage(void)
{
  BeUser();
  system("netscape -remote openURL\\(http://www.ling.ed.ac.uk/~oliphant/grip\\)");
}

void MakeHelpPage(void)
{
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *label;

  helppage=MakeNewPage("Help");

  vbox=gtk_vbox_new(TRUE,2);

  label=gtk_label_new("Helpful information can be found in the README\n"
		      "file that came with the distribution.");
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  label=gtk_label_new("If you would like to submit a bug report,\n"
		      "please ensure that you have first carefully\n"
		      "read both the README and TODO files.");
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  button=gtk_button_new_with_label("Submit Bug Report");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
      GTK_SIGNAL_FUNC(BugReport),NULL);
  gtk_box_pack_start(GTK_BOX(vbox),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  gtk_container_add(GTK_CONTAINER(helppage),vbox);
  gtk_widget_show(vbox);
}

void MakeAboutPage(void)
{
  GtkWidget *vbox,*vbox2,*hbox;
  GtkWidget *label;
  GtkWidget *logo;
  GtkWidget *ebox;
  GtkWidget *button;
  char versionbuf[20];

  aboutpage=MakeNewPage("About");

  ebox=gtk_event_box_new();
  gtk_widget_set_name(ebox,"blackback");

  vbox=gtk_vbox_new(TRUE,5);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

#ifndef GRIPCD
  logo=Loadxpm(grip_xpm);
#else
  logo=Loadxpm(gcd_xpm);
#endif

  gtk_box_pack_start(GTK_BOX(vbox),logo,FALSE,FALSE,0);
  gtk_widget_show(logo);

  vbox2=gtk_vbox_new(TRUE,0);

  sprintf(versionbuf,"Version %s",VERSION);
  label=gtk_label_new(versionbuf);
  gtk_widget_set_name(label,"blackback");
  gtk_box_pack_start(GTK_BOX(vbox2),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  label=gtk_label_new("Copyright (c) 1998-1999, Mike Oliphant");
  gtk_widget_set_name(label,"blackback");
  gtk_box_pack_start(GTK_BOX(vbox2),label,FALSE,FALSE,0);
  gtk_widget_show(label);


  hbox=gtk_hbox_new(TRUE,0);

  button=gtk_button_new_with_label("http://www.ling.ed.ac.uk/~oliphant/grip");
  gtk_widget_set_name(button,"darkergrey");
  gtk_widget_set_name(GTK_BUTTON(button)->child,"blackback");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(Homepage),NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  gtk_box_pack_start(GTK_BOX(vbox2),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);
  

  gtk_container_add(GTK_CONTAINER(vbox),vbox2);
  gtk_widget_show(vbox2);

  gtk_container_add(GTK_CONTAINER(ebox),vbox);
  gtk_widget_show(vbox);

  gtk_container_add(GTK_CONTAINER(aboutpage),ebox);
  gtk_widget_show(ebox);
}

GtkTooltips *MakeToolTip(void)
{
  GtkTooltips *tip;

  tip=gtk_tooltips_new();

  gtk_tooltips_set_delay(tip,1250);

  return tip;
}

GtkWidget *MakeEditBox(void)
{
  GtkWidget *vbox,*hbox;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *frame;
  GtkWidget *item;
  int genre;

  frame=gtk_frame_new(NULL);

  vbox=gtk_vbox_new(FALSE,0);

  hbox=gtk_hbox_new(FALSE,0);

  label=gtk_label_new("Title");
  gtk_widget_set_usize(label,40,0);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  title_edit_entry=gtk_entry_new_with_max_length(256);
  gtk_signal_connect(GTK_OBJECT(title_edit_entry),"changed",
      GTK_SIGNAL_FUNC(TitleEditChanged),NULL);
  gtk_entry_set_position(GTK_ENTRY(title_edit_entry),0);
  gtk_box_pack_start(GTK_BOX(hbox),title_edit_entry,TRUE,TRUE,0);
  gtk_widget_show(title_edit_entry);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);

  hbox=gtk_hbox_new(FALSE,0);

  label=gtk_label_new("Artist");
  gtk_widget_set_usize(label,40,0);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  artist_edit_entry=gtk_entry_new_with_max_length(256);
  gtk_signal_connect(GTK_OBJECT(artist_edit_entry),"changed",
      GTK_SIGNAL_FUNC(ArtistEditChanged),NULL);
  gtk_entry_set_position(GTK_ENTRY(artist_edit_entry),0);
  gtk_box_pack_start(GTK_BOX(hbox),artist_edit_entry,TRUE,TRUE,0);
  gtk_widget_show(artist_edit_entry);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);

  cddb_genre_combo=gtk_combo_new();

  hbox=gtk_hbox_new(FALSE,0);

  label=gtk_label_new("Genre");
  gtk_widget_set_usize(label,40,0);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  for(genre=0;genre<12;genre++) {
    item=gtk_list_item_new_with_label(CDDBGenre(genre));
    cddb_genre_item_list=g_list_append(cddb_genre_item_list,item);
    gtk_signal_connect(GTK_OBJECT(item), "select",
		       GTK_SIGNAL_FUNC(CDDBGenreChanged),NULL);
    gtk_container_add(GTK_CONTAINER(GTK_COMBO(cddb_genre_combo)->list),item);
    gtk_widget_show(item);
  }

  gtk_box_pack_start(GTK_BOX(hbox),cddb_genre_combo,TRUE,TRUE,0);
  gtk_widget_show(cddb_genre_combo);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);

  hbox=gtk_hbox_new(FALSE,0);

  label=gtk_label_new("Track");
  gtk_widget_set_usize(label,40,0);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  track_edit_entry=gtk_entry_new_with_max_length(256);
  gtk_signal_connect(GTK_OBJECT(track_edit_entry),"changed",
		     GTK_SIGNAL_FUNC(TrackEditChanged),NULL);
#ifndef GRIPCD
  gtk_signal_connect(GTK_OBJECT(track_edit_entry),"activate",
		     GTK_SIGNAL_FUNC(CheckDupNames),NULL);
  gtk_signal_connect(GTK_OBJECT(track_edit_entry),"focus_out_event",
		     GTK_SIGNAL_FUNC(CheckDupNames),NULL);
#endif
  gtk_box_pack_start(GTK_BOX(hbox),track_edit_entry,TRUE,TRUE,0);
  gtk_widget_show(track_edit_entry);

  button=ImageButton(save_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
      GTK_SIGNAL_FUNC(SaveDiscInfo),NULL);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Save disc info",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  button=ImageButton(mail_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
      GTK_SIGNAL_FUNC(SubmitEntryCB),NULL);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Submit disc info",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);

  gtk_container_add(GTK_CONTAINER(frame),vbox);
  gtk_widget_show(vbox);

  return frame;
}

GtkWidget *MakePlayOpts(void)
{
  GtkWidget *ebox;
  GtkWidget *hbox;
  GtkWidget *button;

  play_pix[0]=Loadxpm(playnorm_xpm);
  play_pix[1]=Loadxpm(random_xpm);
  play_pix[2]=Loadxpm(playlist_xpm);

  ebox=gtk_event_box_new();
  gtk_widget_set_name(ebox,"blackback");

  hbox=gtk_hbox_new(FALSE,2);

  playlist_entry=gtk_entry_new_with_max_length(256);
  gtk_signal_connect(GTK_OBJECT(playlist_entry),"activate",
		     GTK_SIGNAL_FUNC(PlaylistChanged),NULL);
  gtk_signal_connect(GTK_OBJECT(playlist_entry),"focus_out_event",
		     GTK_SIGNAL_FUNC(PlaylistChanged),NULL);
  gtk_widget_set_sensitive(GTK_WIDGET(playlist_entry),play_mode==PM_PLAYLIST);
  gtk_box_pack_start(GTK_BOX(hbox),playlist_entry,TRUE,TRUE,0);
  gtk_widget_show(playlist_entry);

  play_indicator=NewBlankPixmap();
  CopyPixmap(GTK_PIXMAP(play_pix[0]),GTK_PIXMAP(play_indicator));

  button=gtk_button_new();
  gtk_container_add(GTK_CONTAINER(button),play_indicator);
  gtk_widget_show(play_indicator);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ChangePlayMode),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Rotate play mode",NULL);
  gtk_widget_show(button);

  button=ImageToggleButton(loop_xpm,playloop);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ToggleLoop),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Toggle loop play",NULL);
  gtk_widget_show(button);

  gtk_container_add(GTK_CONTAINER(ebox),hbox);
  gtk_widget_show(hbox);

  return ebox;
}

GtkWidget *MakeControls(void)
{
  GtkWidget *vbox,*vbox2,*vbox3,*hbox,*imagebox,*hbox2;
  GtkWidget *indicator_box;
  GtkWidget *button;
  GtkWidget *ebox,*lcdbox;
  GtkWidget *image;
  GtkObject *adj;
  struct disc_volume vol;
#ifndef GRIPCD
  int mycpu;
#endif

  ebox=gtk_event_box_new();
  gtk_widget_set_name(ebox,"blackback");

  vbox=gtk_vbox_new(FALSE,0);
  gtk_container_border_width(GTK_CONTAINER(vbox),0);

  vbox3=gtk_vbox_new(FALSE,2);
  gtk_container_border_width(GTK_CONTAINER(vbox3),2);

  lcdbox=gtk_event_box_new();
  gtk_widget_set_name(lcdbox,"LCD");

  hbox2=gtk_hbox_new(FALSE,0);

  imagebox=gtk_vbox_new(FALSE,0);

  image=Loadxpm(upleft_xpm);
  gtk_box_pack_start(GTK_BOX(imagebox),image,FALSE,FALSE,0);
  gtk_widget_show(image);

  image=Loadxpm(lowleft_xpm);
  gtk_box_pack_end(GTK_BOX(imagebox),image,FALSE,FALSE,0);
  gtk_widget_show(image);

  gtk_box_pack_start(GTK_BOX(hbox2),imagebox,FALSE,FALSE,0);
  gtk_widget_show(imagebox);
  
  hbox=gtk_hbox_new(TRUE,0);
  gtk_container_border_width(GTK_CONTAINER(hbox),0);

  current_track_label=gtk_label_new("--");
  gtk_box_pack_start(GTK_BOX(hbox),current_track_label,FALSE,FALSE,0);
  gtk_widget_show(current_track_label);

  button=gtk_button_new();
  gtk_widget_set_name(button,"LCD");
#if (GTK_MINOR_VERSION != 0)
  gtk_button_set_relief(GTK_BUTTON(button),GTK_RELIEF_NONE);
#endif
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ChangeTimeMode),NULL);

  play_time_label=gtk_label_new("--:--");
  gtk_container_add(GTK_CONTAINER(button),play_time_label);
  gtk_widget_show(play_time_label);

  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

#ifndef GRIPCD
  rip_pix[0]=Loadxpm(rip0_xpm);
  rip_pix[1]=Loadxpm(rip1_xpm);
  rip_pix[2]=Loadxpm(rip2_xpm);
  rip_pix[3]=Loadxpm(rip3_xpm);

  mp3_pix[0]=Loadxpm(enc0_xpm);
  mp3_pix[1]=Loadxpm(enc1_xpm);
  mp3_pix[2]=Loadxpm(enc2_xpm);
  mp3_pix[3]=Loadxpm(enc3_xpm);
#endif

  cddb_pix[0]=Loadxpm(cddb0_xpm);
  cddb_pix[1]=Loadxpm(cddb1_xpm);

  indicator_box=gtk_hbox_new(TRUE,0);

#ifndef GRIPCD
  rip_indicator=NewBlankPixmap();
#if (GTK_MINOR_VERSION != 0)
  gtk_box_pack_start(GTK_BOX(indicator_box),rip_indicator,TRUE,TRUE,0);
  gtk_widget_show(rip_indicator);
#endif

  for(mycpu=0;mycpu<num_cpu;mycpu++){
    mp3_indicator[mycpu]=NewBlankPixmap();
#if (GTK_MINOR_VERSION != 0)
    gtk_box_pack_start(GTK_BOX(indicator_box),mp3_indicator[mycpu],TRUE,TRUE,0);
#endif
    gtk_widget_show(mp3_indicator[mycpu]);
  }
#endif
  
  cddb_indicator=NewBlankPixmap();
#if (GTK_MINOR_VERSION != 0)
  gtk_box_pack_start(GTK_BOX(indicator_box),cddb_indicator,TRUE,TRUE,0);
  gtk_widget_show(cddb_indicator);
#endif  
  gtk_box_pack_start(GTK_BOX(hbox),indicator_box,TRUE,TRUE,0);
  gtk_widget_show(indicator_box);

  gtk_container_add(GTK_CONTAINER(hbox2),hbox);
  gtk_widget_show(hbox);

  imagebox=gtk_vbox_new(FALSE,0);

  image=Loadxpm(upright_xpm);
  gtk_box_pack_start(GTK_BOX(imagebox),image,FALSE,FALSE,0);
  gtk_widget_show(image);

  image=Loadxpm(lowright_xpm);
  gtk_box_pack_end(GTK_BOX(imagebox),image,FALSE,FALSE,0);
  gtk_widget_show(image);

  gtk_box_pack_start(GTK_BOX(hbox2),imagebox,FALSE,FALSE,0);
  gtk_widget_show(imagebox);
  
  gtk_container_add(GTK_CONTAINER(lcdbox),hbox2);
  gtk_widget_show(hbox2);

  gtk_box_pack_start(GTK_BOX(vbox3),lcdbox,FALSE,FALSE,0);
  gtk_widget_show(lcdbox);

  gtk_box_pack_start(GTK_BOX(vbox),vbox3,FALSE,FALSE,0);
  gtk_widget_show(vbox3);

  adj=gtk_adjustment_new(255.0,0.0,255.0,1.0,1.0,0.0);
  gtk_signal_connect(adj,"value_changed",
		     GTK_SIGNAL_FUNC(SetVolume),NULL);
  volume_control=gtk_hscale_new(GTK_ADJUSTMENT(adj));

  gtk_scale_set_draw_value(GTK_SCALE(volume_control),FALSE);
  gtk_widget_set_usize(volume_control,0,8);
  gtk_widget_set_name(volume_control,"darkgrey");
  gtk_box_pack_start(GTK_BOX(vbox),volume_control,FALSE,FALSE,0);

  CDGetVolume(cd_desc,&vol);
  gtk_adjustment_set_value(GTK_ADJUSTMENT(adj),(vol.vol_front.left+
						vol.vol_front.right)/2);

  if(volvis) gtk_widget_show(volume_control);

  vbox2=gtk_vbox_new(TRUE,0);

  hbox=gtk_hbox_new(TRUE,0);

  button=ImageButton(playpaus_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(PlayTrackCB),NULL);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Play track / Pause play",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(rew_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"pressed",
		     GTK_SIGNAL_FUNC(RewindCB),(gpointer)1);
  gtk_signal_connect(GTK_OBJECT(button),"released",
		     GTK_SIGNAL_FUNC(RewindCB),(gpointer)0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Rewind",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(ff_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"pressed",
		     GTK_SIGNAL_FUNC(FastFwdCB),(gpointer)1);
  gtk_signal_connect(GTK_OBJECT(button),"released",
		     GTK_SIGNAL_FUNC(FastFwdCB),(gpointer)0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "FastForward",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(prevtrk_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(PrevTrack),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Go to previous track",NULL);
  gtk_widget_show(button);

  button=ImageButton(nexttrk_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(NextTrack),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Go to next track",NULL);
  gtk_widget_show(button);

  button=ImageButton(progtrack_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ToggleProg),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Toggle play mode options",NULL);
  gtk_widget_show(button);

  if(changer_slots>1) {
    button=ImageButton(rotate_xpm);
    gtk_widget_set_name(button,"darkergrey");
    gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
		       GTK_SIGNAL_FUNC(NextDisc),0);
    gtk_tooltips_set_tip(MakeToolTip(),button,
			 "Next Disc",NULL);
    gtk_widget_show(button);
  }

  gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,0);
  gtk_widget_show(hbox);

  hbox=gtk_hbox_new(TRUE,0);

  button=ImageButton(stop_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(StopPlay),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Stop play",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(eject_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(EjectDisc),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Eject disc",NULL);
  gtk_widget_show(button);

  button=ImageButton(vol_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ToggleVol),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Toggle Volume Control",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(edit_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ToggleTrackEdit),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Toggle disc editor",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  if(!local_mode) {
    button=ImageButton(cddbwht_xpm);
    gtk_widget_set_name(button,"darkergrey");
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
		       GTK_SIGNAL_FUNC(CDDBToggle),0);
    gtk_tooltips_set_tip(MakeToolTip(),button,
			 "Initiate/abort CDDB lookup",NULL);
    gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
    gtk_widget_show(button);
  }

  button=ImageButton(minmax_xpm);
  gtk_widget_set_name(button,"darkergrey");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(MinMax),0);
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Toggle track display",NULL);
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  button=ImageButton(quit_xpm);
  gtk_widget_set_name(button,"darkergrey");
#ifndef GRIPCD
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Exit Grip",NULL);
#else
  gtk_tooltips_set_tip(MakeToolTip(),button,
		       "Exit GCD",NULL);
#endif
  gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(ShutDown),0);
  gtk_widget_show(button);
  
  gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,0);
  gtk_widget_show(hbox);

  gtk_box_pack_start(GTK_BOX(vbox),vbox2,TRUE,TRUE,0);
  gtk_widget_show(vbox2);


  gtk_container_add(GTK_CONTAINER(ebox),vbox);
  gtk_widget_show(vbox);

  return ebox;
}


GtkWidget *ImageButton(char **xpm)
{
  GtkWidget *button;
  GtkWidget *image;

  button=gtk_button_new();

  image=Loadxpm(xpm);
  gtk_container_add(GTK_CONTAINER(button),image);
  gtk_widget_show(image);

  return button;
}

GtkWidget *ImageToggleButton(char **xpm, gint state)
{
  GtkWidget *button;
  GtkWidget *image;

  button=gtk_toggle_button_new();

#define GTK_VER (GTK_MAJOR_VERSION*100 + GTK_MINOR_VERSION*10 + GTK_MICRO_VERSION)

#if GTK_VER < 120
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), state);
#else
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), state);
#endif

  image=Loadxpm(xpm);
  gtk_container_add(GTK_CONTAINER(button),image);
  gtk_widget_show(image);

  return button;
}

GtkWidget *Loadxpm(char **xpm)
{
  GdkBitmap *mask;
  GtkStyle *style;
  GtkWidget *pixmapwid;
  GdkPixmap *pixmap;

  style=gtk_widget_get_style(window);

  pixmap=gdk_pixmap_create_from_xpm_d(window->window,&mask,
				      &style->bg[GTK_STATE_NORMAL],
				      (gchar **)xpm);

  pixmapwid=gtk_pixmap_new(pixmap,mask);

  return pixmapwid;
}

GtkWidget *NewBlankPixmap(void)
{
  GdkPixmap *gdkpix;
  GdkBitmap *mask;

  gtk_pixmap_get(GTK_PIXMAP(empty_image),&gdkpix,&mask);

  return gtk_pixmap_new(gdkpix,mask);
}

void CopyPixmap(GtkPixmap *src,GtkPixmap *dest)
{
  GdkPixmap *gdkpix;
  GdkBitmap *mask;

  gtk_pixmap_get(src,&gdkpix,&mask);
  gtk_pixmap_set(dest,gdkpix,mask);
}

void Usage(void)
{
#ifndef GRIPCD
  printf("\nUsage: grip [-d <device>] [-s] [-l]\n\n");
#else
  printf("\nUsage: gcd [-d <device>] [-f] [-l]\n\n");
#endif
  printf("  -d <device>        Use <device> as cd-rom device\n");
#ifndef GRIPCD
  printf("  -s                 Launch in \"small\" (cd-only) mode\n");
#else
  printf("  -f                 Launch in \"full\" (track-display) mode\n");
#endif
  printf("  -l                 'local' mode -- don't try to use CDDB\n");
  printf("  -v                 Verbose (debug) mode\n");
  printf("\n");
  printf("See the README file for more information\n\n");

  exit(-1);
}

void ParseArgs(int numargs,char *args[])
{
  int num=1;
  int pos;
  gboolean skipout;

  while(num<numargs) {
    if(*args[num]!='-') Usage();

    for(pos=1,skipout=FALSE;args[num][pos]&&!skipout;pos++) {
      switch(args[num][pos]) {
      case 'h':
      case 'H':
      case '?':
	Usage();
	break;
      case 'd':
	strncpy(cddevice,args[++num],80);
	skipout=TRUE;
	break;
      case 's':
	minimized=1;
	break;
      case 'f':
	minimized=0;
	break;
      case 'l':
	local_mode=TRUE;
	break;
      case 'v':
	do_debug=TRUE;
	break;
      default:
	printf("Unrecognized argument [-%c]\n",args[num][1]);
	Usage();
	break;
      }
    }
    
    num++;
  }
}

void ParseFileFmt(char *instr,char *outstr,int track,int startsec,int endsec)
{
  int left=1024;

  for(;*instr;instr++) {
    if(*instr=='%') {
      switch(*(++instr)) {
      case 't':
	snprintf(outstr,left,"%02d",track);
	break;
      case 'b':
	snprintf(outstr,left,"%d",startsec);
	break;
      case 'e':
	snprintf(outstr,left,"%d",endsec);
	break;
      case 'n':
	if(*ddata.data_track[track-1].track_name)
	  snprintf(outstr,left,"%s",ddata.data_track[track-1].track_name);
	else snprintf(outstr,left,"Track%d",track);
	break;
      case 'a':
	if(*ddata.data_artist)
	  snprintf(outstr,left,"%s",ddata.data_artist);
	else strncpy(outstr,"NoArtist",left);
	break;
      case 'd':
	if(*ddata.data_title)
	  snprintf(outstr,left,"%s",ddata.data_title);
	else strncpy(outstr,"NoTitle",left);
	break;
      case 'i':
	snprintf(outstr,1024,"%02lx",CDDBDiscid(cd_desc));
	break;
      }

      UnderScore(outstr);

      left-=strlen(outstr);
      outstr+=strlen(outstr);
    }
    else {
      if(left) {
	*outstr++=*instr;
	left--;
      }
    }
  }

  *outstr='\0';
}

/* Rip-related functions */

#ifndef GRIPCD

#ifdef MP3DB
void DBScan(void)
{
  int track;

  if(!have_disc) return;

  for(track=0;track<info.disc_totaltracks;track++)
    AddSQLEntry(track);
}

void AddSQLEntry(int track)
{
  int sqlpid;
  char track_str[4];
  char frame_str[11];
  char length_str[11];
  char playtime_str[6];

  snprintf(track_str,4,"%d",track);
  snprintf(frame_str,11,"%d",info.track[track].track_start);
  snprintf(length_str,11,"%d",(info.track[track+1].track_start-1)-
	   info.track[track].track_start);
  snprintf(playtime_str,6,"%d:%d",info.track[track].track_length.minutes,
	   info.track[track].track_length.seconds);

  printf("exe is %s\n",INSTALLDIR "/mp3insert");

  sqlpid=fork();

  if(sqlpid==0) {
    close(ConnectionNumber(GDK_DISPLAY()));
    RedirectIO(FALSE);
    BeUser();
    execl(INSTALLDIR "/mp3insert","mp3insert","-a",ddata.data_artist,
	  "-t",ddata.data_track[track].track_name,"-d",ddata.data_title,
	  "-g",id3_genres[id3_genre_number],"-n",track_str,
	  "-f",frame_str,"-l",length_str,"-m",playtime_str,NULL);

    _exit(0);
  }

  waitpid(sqlpid,NULL,0);

}
#endif /* #ifdef MP3DB */

void UpdateRipProgress(void)
{
  struct stat mystat;
  int quarter;
  gfloat percent;
  int mycpu; 

  static int mp3_quarter[MAX_NUM_CPU];
  static int rip_quarter;
  static int encoded_track[MAX_NUM_CPU];

  if(ripping) {
    if(stat(ripfile,&mystat)>=0) {
      percent=(gfloat)mystat.st_size/(gfloat)ripsize;
      if(percent>1.0) percent=1.0;
    }
    else {
      percent=0;
    }
    
    gtk_progress_bar_update(GTK_PROGRESS_BAR(ripprogbar),percent);
	
    quarter=(int)(percent*4.0);
	
    if(quarter<4&&(quarter!=rip_quarter)) {
      rip_quarter=quarter;
      CopyPixmap(GTK_PIXMAP(rip_pix[rip_quarter]),GTK_PIXMAP(rip_indicator));
    }

    if(waitpid(rippid,NULL,WNOHANG)) {
      waitpid(rippid,NULL,0);
      ripping=0;

      gtk_progress_bar_update(GTK_PROGRESS_BAR(ripprogbar),0.0);
      CopyPixmap(GTK_PIXMAP(empty_image),GTK_PIXMAP(rip_indicator));

      if(!stop_rip) {
	if(doencode&&encoding!=(1<<num_cpu)-1) {
	  quarter=0;
	  for(mycpu=0;mycpu<num_cpu;mycpu++){
	    if(quarter==0&&!(encoding&(1<<mycpu))) {
	      mp3_quarter[mycpu]=0;
	      encoded_track[mycpu]=encode_track=MP3Encode(encode_track,mycpu);
	      quarter=1;
	    }
	  }
	}
	if(!rip_partial&&num_wavs<max_wavs) {
	  RipNextTrack();
	}
      }

      if(!ripping) {
	if(!encoding) stop_rip=0;
	gtk_label_set(GTK_LABEL(rip_prog_label),"Rip: Idle");
      }
    }
  }
  
  for(mycpu=0;mycpu<num_cpu;mycpu++){
    if(encoding&(1<<mycpu)) {
      if(stat(mp3file[mycpu],&mystat)>=0) {
        percent=(gfloat)mystat.st_size/(gfloat)mp3size[mycpu];
        if(percent>1.0) percent=1.0;
      }
      else {
        percent=0;
      } 

      gtk_progress_bar_update(GTK_PROGRESS_BAR(mp3progbar[mycpu]),percent);
        
      quarter=(int)(percent*4.0);
 
      if(quarter<4&&(quarter!=mp3_quarter[mycpu])) {
        mp3_quarter[mycpu]=quarter;
        CopyPixmap(GTK_PIXMAP(mp3_pix[mp3_quarter[mycpu]]),
		   GTK_PIXMAP(mp3_indicator[mycpu]));
      } 
   
      if(waitpid(mp3pid[mycpu],NULL,WNOHANG)) {
        waitpid(mp3pid[mycpu],NULL,0);
        encoding-=(1<<mycpu);

        gtk_progress_bar_update(GTK_PROGRESS_BAR(mp3progbar[mycpu]),0.0);
        CopyPixmap(GTK_PIXMAP(empty_image),GTK_PIXMAP(mp3_indicator[mycpu]));
        mp3_quarter[mycpu]=0;

        if(delete_wavs) unlink(rip_delete_file[mycpu]);

        num_wavs--;

        if(!stop_rip) {
          if(doid3) ID3Add(mp3file[mycpu],encoded_track[mycpu]-1);

#ifdef MP3DB
	  if(add_to_db) AddSQLEntry(encoded_track[mycpu]-1);
#endif
       
          if(!rip_partial) {
            if(!ripping&&num_wavs<max_wavs&&RipNextTrack()) doencode=1;
          }
         
          if(!rip_partial&&doencode){
            encoded_track[mycpu]=encode_track=MP3Encode(encode_track,mycpu);
          }
        }
	else stop_rip=0;

        if(!(encoding&(1<<mycpu)))
	  gtk_label_set(GTK_LABEL(mp3_prog_label[mycpu]),"MP3: Idle");
      }
    }  
  }
}

char *FindRoot(char *str)
{
  char *c;

  for(c=str+strlen(str);c>str;c--) {
    if(*c=='/') return c+1;
  }

  return c;
}

void MakeArgs(char *str,char **args,int maxargs)
{
  int arg;
  gboolean inquotes=FALSE;

  for(arg=0,args[0]=str;*str&&arg<(maxargs-1);str++) {
    if(*str=='"') {
      if(inquotes) *str='\0';
      else args[arg]=str+1;

      inquotes=!inquotes;
    }
    else if(*str==' '&&!inquotes) {
      *str='\0';
      args[++arg]=str+1;
    }
  }

  args[++arg]=NULL;
}

void ParseID3Cmd(char *instr,char *outstr,int track, char *filename)
{
  int left=1024;

  for(;*instr;instr++) {
    if(*instr=='%') {
      switch(*(++instr)) {
      case 't':
	snprintf(outstr,left,"%02d",track);
	break;
      case 'n':
	if(*ddata.data_track[track].track_name)
	  snprintf(outstr,left,"\"%s\"",ddata.data_track[track].track_name);
	else snprintf(outstr,left,"Track%d",track);
	break;
      case 'a':
	if(*ddata.data_artist)
	  snprintf(outstr,left,"\"%s\"",ddata.data_artist);
	else strncpy(outstr,"NoArtist",left);
	break;
      case 'd':
	if(*ddata.data_title)
	  snprintf(outstr,left,"\"%s\"",ddata.data_title);
	else strncpy(outstr,"NoTitle",left);
	break;
      case 'g':
	snprintf(outstr,left,"%d",id3_genre_number);
	break;
      case 'G':
	snprintf(outstr,left,"\"%s\"",id3_genres[id3_genre_number]);
	break;
      case 'f':
	snprintf(outstr,left,"\"%s\"",filename);
	break;
      }

      left-=strlen(outstr);
      outstr+=strlen(outstr);
    }
    else {
      if(left) {
	*outstr++=*instr;
	left--;
      }
    }
  }

  *outstr='\0';
}
void ParseRipCmd(char *instr,char *outstr,int track,int startsec,int endsec,
		 char *filename)
{
  int left=1024;

  for(;*instr;instr++) {
    if(*instr=='%') {
      switch(*(++instr)) {
      case 't':
	snprintf(outstr,left,"%d",track);
	break;
      case 'b':
	snprintf(outstr,left,"%d",startsec);
	break;
      case 'e':
	snprintf(outstr,left,"%d",endsec);
	break;
      case 'f':
	snprintf(outstr,left,"\"%s\"",filename);
	break;
      case 'c':
	strncpy(outstr,cddevice,left);
	break;
      }

      left-=strlen(outstr);
      outstr+=strlen(outstr);
    }
    else {
      if(left) {
	*outstr++=*instr;
	left--;
      }
    }
  }

  *outstr='\0';
}

void ParseMP3Cmd(char *instr,char *outstr,char *ripfile,char *mp3)
{
  int left=1024;

  for(;*instr;instr++) {
    if(*instr=='%') {
      switch(*(++instr)) {
      case 'f':
	snprintf(outstr,left,"\"%s\"",ripfile);
	break;
      case 'o':
	snprintf(outstr,left,"\"%s\"",mp3);
	break;
      }

      left-=strlen(outstr);
      outstr+=strlen(outstr);
    }
    else {
      if(left) {
	*outstr++=*instr;
	left--;
      }
    }
  }

  *outstr='\0';
}

void KillChild(void)
{
  int mycpu;
  
  if(!(ripping||encoding)) return;

  stop_rip=1;

  if(ripping) {
    kill(rippid,SIGKILL);
    doencode=0;
  }
  
  for(mycpu=0;mycpu<num_cpu;mycpu++){
    if(encoding&(1<<mycpu)) kill(mp3pid[mycpu],SIGKILL);
  }
}

void RedirectIO(gboolean redirect_output)
{
  int fd;

  fd=open("/dev/null",O_RDWR);
  dup2(fd,0);

  if(redirect_output) {
    dup2(fd,1);
    dup2(fd,2);
  }

  close(fd); 
  close(cd_desc);
}

void ID3Add(char *file, int track)
{
  char parsedstr[1024];
  char *args[20];
  pid_t id3pid;

  printf("add ID3 to %s\n", file);

  ParseID3Cmd(id3cmdline,parsedstr,track,file);
    
  Debug("cmdline is [%s]\n",parsedstr);

  args[0]=FindRoot(id3exename);

  MakeArgs(parsedstr,args+1,19);
    
  id3pid=fork();

  if(id3pid==0) {
    close(ConnectionNumber(GDK_DISPLAY()));
    RedirectIO(TRUE);
    BeUser();
    execv(id3exename,args);
    _exit(0);
  }

  waitpid(id3pid,NULL,0);
}

int MP3Encode(int track,int cpu)
{
  char parsedstr[1024];
  char tmp[1024];
  char *args[20];
  int startsec,endsec;

  if(!rip_partial)
    for(;(track<info.disc_totaltracks)&&
          !TrackIsChecked(track)&&
          (!ripping||(track<(rip_track-1)));track++);

  if((ripping&&(track==(rip_track-1)))||
     (track==info.disc_totaltracks)) return 0;

  if(have_disc&&encode_track>=0) {
    encoding+=1<<cpu;

    ToggleChecked(track);

    CopyPixmap(GTK_PIXMAP(mp3_pix[0]),GTK_PIXMAP(mp3_indicator[cpu]));

    sprintf(tmp,"MP3: Trk %d",track+1);
    gtk_label_set(GTK_LABEL(mp3_prog_label[cpu]),tmp);

    printf("MP3 track %d\n",track+1);

    if(rip_partial) {
      startsec=start_sector;
      endsec=end_sector;
    }
    else {
      startsec=0;
      endsec=(info.track[track+1].track_start-1)-
	info.track[track].track_start;
    }

    ParseFileFmt(outputdir,parsedstr,track+1,startsec,endsec);

    MakePath(parsedstr);

    ParseFileFmt(mp3fileformat,tmp,track+1,startsec,endsec);

    sprintf(mp3file[cpu],"%s%s",parsedstr,tmp);

    MakeDirs(mp3file[cpu]);

    printf("%i: Encoding to %s\n",cpu+1,mp3file[cpu]);

    unlink(mp3file[cpu]);

    mp3size[cpu]=(((endsec-startsec)+1)*2352)/11;

    /* Construct the .wav filename to encode */

    ParseFileFmt(outputdir,parsedstr,track+1,startsec,endsec);

    MakePath(parsedstr);

    ParseFileFmt(ripfileformat,tmp,track+1,startsec,endsec);

    sprintf(rip_delete_file[cpu],"%s%s",parsedstr,tmp);

    ParseMP3Cmd(mp3cmdline,parsedstr,rip_delete_file[cpu],mp3file[cpu]);

    args[0]=FindRoot(mp3exename);

    MakeArgs(parsedstr,args+1,19);

    mp3pid[cpu]=fork();

    if(mp3pid[cpu]==0) {
      close(ConnectionNumber(GDK_DISPLAY()));
      RedirectIO(TRUE);
      BeUser();
      setsid();
      nice(mp3nice);
      execv(mp3exename,args);
      _exit(0);
    }

    track++;
  }
  return track;
}

void DoRip(GtkWidget *widget,gpointer data)
{
  gboolean result;

  if(!FileExists(ripexename)) {
    BoolDialog("Invalid rip executable\nCheck your rip config","OK",
	       NULL,NULL,NULL);

    return;
  }

  if(data&&!FileExists(mp3exename)) {
    BoolDialog("Invalid MP3 executable\nCheck your MP3 config","OK",
	       NULL,NULL,NULL);

    return;
  }

  num_wavs=0;

  if(ripping|encoding) return;

  if(rip_partial)
    encode_track=rip_track=CURRENT_TRACK;
  else {
    setCurrentTrackIndex(0);
    encode_track=rip_track=0;
  }

  if(data) doencode=1;
  else doencode=0;

  result=RipNextTrack();
  if(!result) {
    doencode=0;
    BoolDialog("No tracks selected","OK",NULL,NULL,NULL);
  }
}

void MakeDirs(char *path)
{
  char dir[256];
  char *s;
  int len;

  BeUser();

  for(len=0,s=path;*s;s++,len++) {
    if(*s=='/') {
      strncpy(dir,path,len);
      dir[len]='\0';

      if(!FileExists(dir))
	mkdir(dir,0755);
    }
  }

  BeRoot();
}

char *MakePath(char *str)
{
  int len;

  len=strlen(str)-1;

  if(str[len]!='/') {
    str[len+1]='/';
    str[len+2]='\0';
  }

  return str;
}

gboolean RipNextTrack(void)
{
  char parsedstr[1024];
  char tmp[1024];
  char *args[20];
  int startsec,endsec;

  if(ripping) return FALSE;

  if(!rip_partial)
    for(;(rip_track<info.disc_totaltracks)&&
	  !TrackIsChecked(rip_track);rip_track++);

  if(rip_track==info.disc_totaltracks)  /* Finished ripping */
    return FALSE;

  /* We have a track to rip */

  if(have_disc&&rip_track>=0) {
    ripping=1;
    rip_quarter=0;

    if(!rip_partial){
      gtk_clist_select_row(GTK_CLIST(trackclist),rip_track,0);		
      if(!doencode) ToggleChecked(rip_track);
    }

    CopyPixmap(GTK_PIXMAP(rip_pix[0]),GTK_PIXMAP(rip_indicator));

    sprintf(tmp,"Rip: Trk %d",rip_track+1);
    gtk_label_set(GTK_LABEL(rip_prog_label),tmp);

    CDStop(cd_desc);

    if(rip_partial) {
      startsec=start_sector;
      endsec=end_sector;
    }
    else {
      startsec=0;
      endsec=(info.track[rip_track+1].track_start-1)-
	info.track[rip_track].track_start;
    }

    /*    endsec=100;*/

    ParseFileFmt(outputdir,parsedstr,rip_track+1,startsec,endsec);

    MakePath(parsedstr);

    ripsize=44+((endsec-startsec)+1)*2352;
    ParseFileFmt(ripfileformat,tmp,rip_track+1,startsec,endsec);

    sprintf(ripfile,"%s%s",parsedstr,tmp);

    MakeDirs(ripfile);

    printf("Ripping track %d to %s\n",rip_track+1,ripfile);

    unlink(ripfile);

    ParseRipCmd(ripcmdline,parsedstr,rip_track+1,startsec,endsec,
		ripfile);

    Debug("Rip executable is [%s]\n",ripexename);
    Debug("cmdline is [%s]\n",parsedstr);

    args[0]=FindRoot(ripexename);

    MakeArgs(parsedstr,args+1,19);

    rippid=fork();

    if(rippid==0) {
      close(ConnectionNumber(GDK_DISPLAY()));
      //      RedirectIO(TRUE);
      nice(ripnice);
      execv(ripexename,args);

      Debug("Exec failed\n");
      _exit(0);
    }

    num_wavs++;
    rip_track++;

    return TRUE;
  }
  else return FALSE;
}

void PlaySegmentCB(GtkWidget *widget,gpointer data)
{
  PlaySegment(CURRENT_TRACK);
}

void PlaySegment(int track)
{
  CDPlayFrames(cd_desc,info.track[track].track_start+start_sector,
	       info.track[track].track_start+end_sector);
}

void ID3GenreChanged(GtkWidget *widget,gint number)
{
  id3_genre_number = number;
}

GtkWidget *MakeRangeSelects(void)
{
  GtkWidget *vbox;
  GtkWidget *numentry;

  vbox=gtk_vbox_new(FALSE,0);

  numentry=MakeNumEntry("Start sector",0,5);
  gtk_box_pack_start(GTK_BOX(vbox),numentry,FALSE,FALSE,0);
  gtk_widget_show(numentry);
  
  numentry=MakeNumEntry("End sector",1,5);
  gtk_box_pack_start(GTK_BOX(vbox),numentry,FALSE,FALSE,0);
  gtk_widget_show(numentry);
  
  return vbox;
}

void MakeRipPage(void)
{
  char buf[256];
  GtkWidget *rangesel;
  GtkWidget *vbox,*vbox2,*hbox,*hbox2;
  GtkWidget *button;
  GtkWidget *hsep;
  GtkWidget *check;
  GtkWidget *item;
  GtkWidget *label;
  gchar **id3_genre;
  gint id3_genre_count;
  int mycpu;

  rippage=MakeNewPage("Rip");

  vbox=gtk_vbox_new(FALSE,2);
  gtk_container_border_width(GTK_CONTAINER(vbox),3);

  snprintf(buf,256,"Track: %s",ddata.data_track[CURRENT_TRACK].track_name);
  riptracknamelabel=gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(vbox),riptracknamelabel,FALSE,FALSE,0);
  gtk_widget_show(riptracknamelabel);

  snprintf(buf,256,"Playing time: %d:%02d",
	   info.track[CURRENT_TRACK].track_length.minutes,
	   info.track[CURRENT_TRACK].track_length.seconds);
  riptracktimelabel=gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(vbox),riptracktimelabel,FALSE,FALSE,0);
  gtk_widget_show(riptracktimelabel);

  hbox=gtk_hbox_new(FALSE,3);

  label=gtk_label_new("ID3 Genre:");
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  id3_genre_combo=gtk_combo_new();

  for(id3_genre=id3_genres,id3_genre_count=0;*id3_genre;id3_genre++,
	id3_genre_count++) {
    item = gtk_list_item_new_with_label(*id3_genre);
    id3_genre_item_list = g_list_append(id3_genre_item_list, item);
    gtk_signal_connect(GTK_OBJECT(item), "select",
		       GTK_SIGNAL_FUNC(ID3GenreChanged),
		       (gpointer)id3_genre_count);
    gtk_container_add(GTK_CONTAINER(GTK_COMBO(id3_genre_combo)->list), item);
    gtk_widget_show(item);
  }

  gtk_box_pack_start(GTK_BOX(hbox),id3_genre_combo,TRUE,TRUE,0);
  gtk_widget_show(id3_genre_combo);

  item = GTK_WIDGET(g_list_nth(id3_genre_item_list,id3_genre_number)->data);
  gtk_list_select_child (GTK_LIST (GTK_COMBO(id3_genre_combo)->list), item);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);


  hsep=gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox),hsep,TRUE,TRUE,0);
  gtk_widget_show(hsep);

  hbox=gtk_hbox_new(FALSE,5);

  vbox2=gtk_vbox_new(FALSE,3);

  button=gtk_button_new_with_label("Rip");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(DoRip),(gpointer)0);
  gtk_box_pack_start(GTK_BOX(vbox2),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  button=gtk_button_new_with_label("MP3-encode");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(DoRip),(gpointer)1);
  gtk_box_pack_start(GTK_BOX(vbox2),button,FALSE,FALSE,0);
  gtk_widget_show(button);

  button=gtk_button_new_with_label("Abort execution");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(KillChild),NULL);
  gtk_box_pack_start(GTK_BOX(vbox2),button,FALSE,FALSE,0);
  gtk_widget_show(button);

#ifdef MP3DB
  button=gtk_button_new_with_label("MP3 DB Scan");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(DBScan),NULL);
  gtk_box_pack_start(GTK_BOX(vbox2),button,FALSE,FALSE,0);
  gtk_widget_show(button);
#endif

  gtk_box_pack_start(GTK_BOX(hbox),vbox2,FALSE,FALSE,0);
  gtk_widget_show(vbox2);

  vbox2=gtk_vbox_new(FALSE,3);

  hbox2=gtk_hbox_new(FALSE,5);

  check=MakeCheckButton("Rip partial track",0);
  gtk_box_pack_start(GTK_BOX(hbox2),check,FALSE,FALSE,0);
  gtk_widget_show(check);

  button=gtk_button_new_with_label("Play");
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(PlaySegmentCB),
		     (gpointer)NULL);
  gtk_box_pack_start(GTK_BOX(hbox2),button,TRUE,TRUE,0);
  gtk_widget_show(button);

  gtk_box_pack_start(GTK_BOX(vbox2),hbox2,FALSE,FALSE,0);
  gtk_widget_show(hbox2);

  play_sector_label=gtk_label_new("Current sector:      0");
  gtk_box_pack_start(GTK_BOX(vbox2),play_sector_label,FALSE,FALSE,0);
  gtk_widget_show(play_sector_label);

  rangesel=MakeRangeSelects();
  gtk_box_pack_start(GTK_BOX(vbox2),rangesel,FALSE,FALSE,0);
  gtk_widget_show(rangesel);

  gtk_box_pack_start(GTK_BOX(hbox),vbox2,TRUE,TRUE,0);
  gtk_widget_show(vbox2);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,TRUE,TRUE,0);
  gtk_widget_show(hbox);

  hsep=gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox),hsep,TRUE,TRUE,0);
  gtk_widget_show(hsep);

  hbox=gtk_hbox_new(FALSE,5);

  rip_prog_label=gtk_label_new("Rip: Idle");
  gtk_widget_set_usize(rip_prog_label,WINWIDTH-215,0);
  gtk_box_pack_start(GTK_BOX(hbox),rip_prog_label,FALSE,FALSE,0);
  gtk_widget_show(rip_prog_label);

  ripprogbar=gtk_progress_bar_new();
  gtk_widget_set_usize(ripprogbar,WINWIDTH-95,0);
  gtk_box_pack_start(GTK_BOX(hbox),ripprogbar,FALSE,FALSE,0);
  gtk_widget_show(ripprogbar);

  gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
  gtk_widget_show(hbox);

  for(mycpu=0;mycpu<num_cpu;mycpu++){
    hbox=gtk_hbox_new(FALSE,5);

    mp3_prog_label[mycpu]=gtk_label_new("MP3: Idle");
    gtk_widget_set_usize(mp3_prog_label[mycpu],WINWIDTH-215,0);
    gtk_box_pack_start(GTK_BOX(hbox),mp3_prog_label[mycpu],FALSE,FALSE,0);
    gtk_widget_show(mp3_prog_label[mycpu]);

    mp3progbar[mycpu]=gtk_progress_bar_new();
    gtk_widget_set_usize(mp3progbar[mycpu],WINWIDTH-95,0);
    gtk_box_pack_start(GTK_BOX(hbox),mp3progbar[mycpu],FALSE,FALSE,0);
    gtk_widget_show(mp3progbar[mycpu]);

    gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
    gtk_widget_show(hbox);
  }

  gtk_container_add(GTK_CONTAINER(rippage),vbox);
  gtk_widget_show(vbox);
}

#endif

int main(int argc,char *argv[])
{
  char buf[256];

  srand(time(NULL));

  ParseArgs(argc,argv);

  DoLoadConfig();

  if(getuid()==0) root=1;

  cd_desc=CDInitDevice(cddevice);
  if(cd_desc==-1) {
    printf("Error: Unable to open %s\n",cddevice);
    exit(0);
  }

  gtk_init(&argc,&argv);

  if((GTK_MINOR_VERSION!=gtk_minor_version)) {
    printf("Warning: This version of Grip was compiled using gtk+ "
	   "version %d.%d.%d, and you are running gtk+ version %d.%d.%d...\n",
	   GTK_MAJOR_VERSION,GTK_MINOR_VERSION,GTK_MICRO_VERSION,
	   gtk_major_version,gtk_minor_version,gtk_micro_version);
  }

  if(!no_interrupt)
    CDStop(cd_desc);
    
  changer_slots=CDChangerSlots(cd_desc);

#ifndef GRIPCD
  snprintf(buf,256,"%s/grip.rc",AUXDIR);
#else
  snprintf(buf,256,"%s/gcd.rc",AUXDIR);
#endif
  gtk_rc_parse(buf);

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_realize(window);

  gtk_window_set_policy(GTK_WINDOW(window),FALSE,TRUE,FALSE);

  sprintf(buf,"%s %s",PROGRAM,VERSION);
  gtk_window_set_title(GTK_WINDOW(window),buf);

  gtk_signal_connect(GTK_OBJECT(window),"delete_event",
		     GTK_SIGNAL_FUNC(ShutDown),NULL);

  notebook=gtk_notebook_new();
  gtk_widget_set_usize(notebook,WINWIDTH,0);

  winbox=gtk_vbox_new(FALSE,3);
  if(!minimized) gtk_container_border_width(GTK_CONTAINER(winbox),3);

  MakeTrackPage();

#ifndef GRIPCD
  MakeRipPage();
#endif

  MakeConfigPage();

  gtk_box_pack_start(GTK_BOX(winbox),notebook,TRUE,TRUE,0);
  if(!minimized) gtk_widget_show(notebook);

  gtk_container_add(GTK_CONTAINER(window),winbox);
  gtk_widget_show(winbox);

  wht_check_image=Loadxpm(whtcheck_xpm);
  blk_check_image=Loadxpm(blkcheck_xpm);
  empty_image=Loadxpm(empty_xpm);

  MakeHelpPage();
  MakeAboutPage();

  gdk_color_white(gdk_window_get_colormap(window->window),&gdkwhite);
  gdk_color_black(gdk_window_get_colormap(window->window),&gdkblack);

  track_edit_box=MakeEditBox();

  gtk_box_pack_start(GTK_BOX(winbox),track_edit_box,FALSE,FALSE,0);

  playopts=MakePlayOpts();
  gtk_box_pack_start(GTK_BOX(winbox),playopts,FALSE,FALSE,0);

  controls=MakeControls();
  gtk_box_pack_start(GTK_BOX(winbox),controls,FALSE,FALSE,0);
  gtk_widget_show(controls);

  CheckNewDisc();
  UpdateDisplay();

  gtk_widget_show(window);

  /* Counteract the oops I made in 1.6 */

  if(*dbserver.cgi_prog=='/')
    BoolDialog("You should probably remove the leading\n"
	       "'/' from your primary server's CGI path.\n",
	       "OK",NULL,NULL,NULL);

  wait_cursor=gdk_cursor_new(GDK_WATCH);

  gtk_timeout_add(1000,TimeOut,0);

  gtk_main();

  if(!no_interrupt)
    CDStop(cd_desc);

  return 0;
}
